package org.gaugebook;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Transparency;
import java.awt.geom.Arc2D;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.VolatileImage;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.sound.sampled.Clip;

import org.gaugebook.math.Angle;

/**
 * AbstractClockGauge
 * 
 * This is the base class for all simple, round clock gauges. It provides fields and methods to easily create
 * your own gauge written in Java. For more information, have a look at the tutorial on the gaugebook website.
 * 
 * @see http://www.gaugebook.org/en/java-gauge-programmieren
 * @author Michael Nagler
 */
public abstract class AbstractClockGauge extends AbstractGauge {

	private static final long serialVersionUID = 1L;

    protected int lastWidth  = 0, lastbgw = 0;
    protected int lastHeigth = 0, lastbgh = 0;
    
    protected VolatileImage vi;
    protected boolean update = false; 

    protected Color scaleLineColor     		= new Color(142,142,142);    // scale line color
    protected Color needleCircleColor  		= new Color(102,102,102);    // needle circle color
    protected Color needleColor       		= new Color(250,250,250); // Color( 224, 142, 48 );  // needle color
    protected Color backgroundColor   		= new Color(80,85,90);
	protected Color valueColor = new Color( 200, 100, 100); 
	private   Color backgroundBorderColor   = Color.BLACK;				// color of outer border 
    protected Color majorTickColor     		= new Color(200,200,200); // Color.WHITE;
    protected Color textColor         		= new Color(200,200,200);
    protected Color minorTickColor = new Color(180,180,180); // Color.WHITE; 
	protected Color valueBackgroundColor = new Color( 40, 40, 40 );
    
    private String text = "";
    protected int decPlaceCount = 0; //[0..3]
    protected boolean showValue = true;
    private NumberFormat format = NumberFormat.getIntegerInstance();
    private Point2D textPosition = new Point2D.Double( 0, -0.3);
    private double textHeight = 1;
	public enum Direction { CLOCKWISE, ANTICLOCKWISE };
	
	public AbstractClockGauge() {
		super();
		setBackground( backgroundColor );
	}
	
	/**
     * conforms the bean properties.
     */
    protected void conformProperties() {
        // minValue, maxValue 
        double min = Math.min( minValue, maxValue ); 
        double max = Math.max( minValue, maxValue ); 
        
        minValue = min;
        maxValue = max;
        // value - sets value to minvalue if value is outside of range.
        if( value < minValue || value > maxValue ) { 
            value = (value < minValue ? minValue : maxValue);
        }
        //majorTickStep
        if(majorTickStep >= (maxValue - minValue)){
            majorTickStep = (maxValue - minValue)/10.0;
        }
        // minorTickStep
        if(minorTickStep >= majorTickStep){
            minorTickStep = majorTickStep/4.0;
        }
        
        // angleExtend
        if(angleExtend < 90.0 || angleExtend > 360.0){
            angleExtend = angleExtend < 90.0 ? 90.0 : 360.0;
        }

        if(angleStart < - 180.0 || angleStart > 180.0){
            angleStart = angleStart < -180.0 ? -180.0 : 180.0;
        }
    }

    public void update() {
    	repaint();
    }
    
    /**
     * Render the component (clock gauge)
     */
	public void paint( Graphics g ) {
        int w = getWidth();
        int h = getHeight();
    	if( g == null || w == 0 || h == 0 ) return;
        
        do {
            double r =  Math.min( w, h ) / 2 - 5;            // radius
            if( r < 0 ) r = 100;
            Point2D cp = new Point2D.Double( w / 2, h / 2 ); // center point

            GraphicsConfiguration gc = this.getGraphicsConfiguration();
        	if( w != lastWidth || h != lastHeigth || vi == null || vi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
        		vi = gc.createCompatibleVolatileImage( w, h, Transparency.OPAQUE ); // gc.getColorModel().getTransparency() );
                lastWidth = w;
                lastHeigth = h;
        	}
        	
        	Graphics2D g2 = vi.createGraphics();
            g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));

           	drawBackground( r, cp, g2 );

           	if( initdone ) {
	            drawText(r,cp,g2);
	            double r0 = r * .55;
	            double r1 = r * .58;
	            double majorTickSize = r * .04;
	            double minorTickSize = r * .02;
	            double majorTickR =  r1 + 4; // major tick radius 
	            double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
	            double needleR = majorTickR + 1.5 * majorTickSize;
	            drawValue(r,needleR, cp,g2);
	            drawScale(r,cp,g2);
	            drawNeedle( r, needleR, cp, g2 );
           	}
            g2.dispose();
        } while( vi.contentsLost() );

        g.drawImage( vi, 0, 0, this );
    }
    
    protected void drawBackground( final double r, final Point2D cp, Graphics2D g2 ) {
    	int w = getWidth();
    	int h = getHeight();
    	boolean redo = false;
		// volatile image
        do {
    		GraphicsConfiguration gc = getGraphicsConfiguration();
        	if( backVi == null || w != lastbgw || h != lastbgh || backVi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
        		backVi = gc.createCompatibleVolatileImage( w, h, Transparency.OPAQUE );
        		lastbgw = w;
        		lastbgh = h;
        	}
	        	Graphics2D bg2 = backVi.createGraphics();
	            bg2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));
	            Shape oldClip = g2.getClip();
	
		        bg2.setColor( getBackgroundColor() );
		        bg2.fillRect(0,0,w,h);
		        
	    		Ellipse2D.Double outer = new Ellipse2D.Double( cp.getX() - r, cp.getY() - r, 2 * r, 2 * r );
		        double r0 = r - r * border_width;
		        Ellipse2D.Double inner = new Ellipse2D.Double( cp.getX() - r0, cp.getY()- r0, 2 * r0, 2 * r0 );
		        
		        Point2D op = getPoint(cp, r , -45);
		        GradientPaint gray2darkgray = new GradientPaint(cp, Color.GRAY, op, Color.DARK_GRAY, true);
		        GradientPaint darkgray2black = new GradientPaint(cp, Color.BLACK, op, Color.DARK_GRAY, true);
		        op = getPoint(cp, r * .5, -45);
		        RoundGradientPaint round_darkgray2black = new RoundGradientPaint(cp, Color.DARK_GRAY, op, Color.BLACK );
	
		        Point2D tr = new Point2D.Double( cp.getX() - r, cp.getY() - r );
		        Point2D tl = new Point2D.Double( cp.getX() + r, cp.getY() - r );
		        Point2D br = new Point2D.Double( cp.getX() - r, cp.getY() + r );
		        Point2D bl = new Point2D.Double( cp.getX() + r, cp.getY() + r );
		        // lower outer ell gradient fill
		        bg2.setPaint(gray2darkgray);
		        Polygon pol = new Polygon();
		        pol.addPoint((int)tr.getX(), (int)tr.getY());
		        pol.addPoint((int)bl.getX(), (int)bl.getY());
		        pol.addPoint((int)br.getX(), (int)br.getY());
		        Area a = new Area( pol );
		        bg2.setClip( a );
		        bg2.fill(outer);
		        bg2.setClip(oldClip);
		        
		        // upper outer ell gradient fill
		        bg2.setPaint(darkgray2black);
		        pol = new Polygon();
		        pol.addPoint((int)tr.getX(), (int)tr.getY());
		        pol.addPoint((int)tl.getX(), (int)tl.getY());
		        pol.addPoint((int)bl.getX(), (int)bl.getY());
		        a = new Area( pol );
		        bg2.setClip( a );
		        bg2.fill(outer);
		        
		        bg2.setClip(oldClip);
		        
		        // inner gradient fill
		        bg2.setPaint(round_darkgray2black );
		        bg2.fill(inner);
	        
		        bg2.dispose();
        } while( backVi.contentsLost() );
    	g2.drawImage( backVi, 0, 0, this );

    }    

    
    /**
     * Draws the scale.
     * 
     * @param r the Radius of Gauge
     * @param cp the center Point of Gauge
     * @param g2 the graphics device
     */
    protected void drawScale( final double r, final Point2D cp, Graphics2D g2, Direction direction ) {
        double r0 = proz(r,65); // 55
        double r1 = proz(r,80); // 58
        
        double majorTickSize = proz(r,7.5);
        double minorTickSize = proz(r,5.5);
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
            g2.setColor( minorTickColor );
            g2.setStroke( new BasicStroke(1.f) );            
            double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            for(double i= angleStart + angleExtend; i >= angleStart ; i-=minorGradSteps){
                Point2D p0 = getPoint( cp, minorTickR, i );
                Point2D p1 = getPoint( cp, minorTickR+minorTickSize, i );
                g2.draw( new Line2D.Double(p0,p1) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
            	double myv = direction == Direction.CLOCKWISE ? v : maxValue - v + minValue; 
            			
                double currentAngle = angleStart + angleExtend - stepAngle * (v - minValue) / majorTickStep;
                Point2D p0 = getPoint( cp, majorTickR, currentAngle );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s = valueToString( myv );
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle ); // marker center point
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
                
            }
            
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }
    }
    
    protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
    	drawScale( r, cp, g2, Direction.CLOCKWISE );
    }

    /**
     * converts value to String.
     * 
     * @param v the value that is to converted.
     * 
     * @return the string representation of value
     */
    protected String valueToString(double v){
        return format.format( v );
    }
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
        //draw current value
        if(showValue){
            g2.setColor( valueBackgroundColor  );
            String s = valueToString( value );
            // center point of current value area
            double rv = proz( r, 45 );  
            Point2D p = getPoint( cp, rv, angleStart + angleExtend + (360 - angleExtend)/2 ); // value center point        
            double h = proz(r,16);   // height of curent value area 
            g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h * 0.7)  ) );
            
            Dimension size = getStringSize( s, g2 );
            int l = s.length();
            double f = l < 3 ? 2.0 : Math.max(2.0 - l*0.1, 1.1);
            double w  = size.getWidth() * f; // width of current value area 
            double corner = Math.min( 10, Math.max( h, w ) / 4.0); // maximum 10
            
            RoundRectangle2D cvArea = new RoundRectangle2D.Double(p.getX()- w/2,p.getY()-h/2,w,h,corner,corner);
            g2.fill( cvArea );
            
            Stroke old = g2.getStroke();
            g2.setColor( scaleLineColor );
            g2.setStroke( new BasicStroke(h < 20 ? 1f : (h < 40 ? 2f : 3f)) );
            g2.draw( cvArea );
            
            g2.setStroke( old );
            
			g2.setColor( valueColor );
            
            double x0 = cvArea.getCenterX() - size.getWidth()/2.0;
            double y0 = cvArea.getCenterY() + size.getHeight()/2.0;
            g2.drawString( s,(float)x0,(float)y0);
        }
    }
    
    /**
     * Draws the needle and current value.
     * 
     * @param r the Radius of Gauge
     * @param needleR the Radius of needle
     * @param cp the center Point of Gauge
     * @param g2 the graphics device
     */
    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2, double value, Direction direction ) {
        double myvalue = direction == Direction.CLOCKWISE ? value : maxValue - value + minValue;

        Angle angleCurrent = new Angle( angleStart + angleExtend * (1.0 - (myvalue - minValue)/(maxValue - minValue)));
        Point2D apL = getPoint( cp, needleR*1.2, angleCurrent.plus(-1));
        Point2D ap = getPoint( cp, needleR*1.3, angleCurrent.getDegrees() );
        Point2D apR = getPoint( cp, needleR*1.2, angleCurrent.plus(1));
        
        Angle angleLeft = new Angle( angleCurrent.getDegrees() -45 );
        Angle  angleRight = new Angle( angleCurrent.getDegrees() +45 );

        Point2D pointLeft  = getPoint( cp, needleR * .03, angleLeft.getDegrees() );
        Point2D pointRight = getPoint( cp, needleR * .03, angleRight.getDegrees() );
        
        Polygon pol = new Polygon();
        pol.addPoint( (int)Math.round(pointLeft.getX()), (int)Math.round(pointLeft.getY()) );
        pol.addPoint( (int)Math.round(apL.getX()), (int)Math.round(apL.getY()) );
        pol.addPoint( (int)Math.round(ap.getX()), (int)Math.round(ap.getY()) );
        pol.addPoint( (int)Math.round(apR.getX()), (int)Math.round(apR.getY()) );
        pol.addPoint( (int)Math.round(pointRight.getX()), (int)Math.round(pointRight.getY()) );
        
        g2.setColor( needleColor );
        g2.fill( pol);

        pointLeft  = getPoint( cp, needleR * .12, angleLeft.plus(-180 +20));
        pointRight = getPoint( cp, needleR * .12, angleLeft.plus(-90 -20));

        apL = getPoint( cp, needleR * .12, angleLeft.plus(+20));
        apR = getPoint( cp, needleR * .12, angleLeft.plus(+90 -20));

        pol = new Polygon();

        pol.addPoint( (int)Math.round(pointLeft.getX()), (int)Math.round(pointLeft.getY()) );
        pol.addPoint( (int)Math.round(pointRight.getX()), (int)Math.round(pointRight.getY()) );
        pol.addPoint( (int)Math.round(apL.getX()), (int)Math.round(apL.getY()) );
        pol.addPoint( (int)Math.round(apR.getX()), (int)Math.round(apR.getY()) );
        
        g2.setColor( Color.GRAY );
        g2.fill( pol);

    }
    
    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    	drawNeedle( r, needleR, cp, g2, value, Direction.CLOCKWISE );
    }

    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2, double value ) {
    	drawNeedle( r, needleR, cp, g2, value, Direction.CLOCKWISE );
    }
    
    /**
     * Draws the text.
     * 
     * @param r the Radius of Gauge
     * @param cp the center Point of Gauge
     * @param g2 the graphics device
     */
    protected void drawText( final double r, final Point2D cp, Graphics2D g2, Point2D position, double textHeight, String text ) {
    	if( text == null || text.length() == 0 ) return;
    	
        double r0 = proz( r, 90 );  
        double r1 = proz( r, 100 ); 
        double h = ( r1 - r0 ) * textHeight;   // height of text 

        g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h ) ) );
        g2.setColor( textColor );
        
        Dimension dim = getStringSize( text, g2 );
        double x0 = cp.getX() - dim.getWidth()/2;
        double y0 = cp.getY() - dim.getHeight()/2;

        x0 = x0 + r * position.getX();
        y0 = y0 + r * position.getY();
        
       	g2.drawString( text,(float)x0,(float)y0);
    }
        
    protected void drawText( final double r, final Point2D cp, Graphics2D g2, Point2D position ) {
    	drawText( r, cp, g2, textPosition, textHeight, getText() );
    }

    protected void drawText( final double r, final Point2D cp, Graphics2D g2 ) {
    	drawText( r, cp, g2, textPosition, textHeight, getText() );
    }

    /**
     * Reloads the buffer for the given two Colors.
     * 
     * @param newColor the new color. 
    private void reloadBuffer(Color newColor, Color newBorderColor){
        // reload the buffer
        int r = newColor.getRed();
        int g = newColor.getGreen();
        int b = newColor.getBlue();
        int a = newColor.getAlpha();
        
        int r2 = newBorderColor.getRed();
        int g2 = newBorderColor.getGreen();
        int b2 = newBorderColor.getBlue();
        int a2 = newBorderColor.getAlpha();
        
        int innerBorderLength = backgroundBuffer.length - BORDER_RGB.length;
        
        for( int i = 0; i < backgroundBuffer.length; i++ ) {
            Color c = new Color( ALL_RGB[i] );
            int r1 = Math.min( 255, c.getRed()  + (i < innerBorderLength ? r : r2) );
            int g1 = Math.min( 255, c.getGreen()+ (i < innerBorderLength ? g : g2) );
            int b1 = Math.min( 255, c.getBlue() + (i < innerBorderLength ? b : b2) );
            int a1 = (i < innerBorderLength ? a : a2);
            backgroundBuffer[i] = new Color( r1, g1, b1, a1 ).getRGB();
        }

    }
     */

    public void drawMarker(double value, Color color, float size, double r, double r0, double r1, Point2D cp, Graphics2D g2 ) {
    	drawMarker(value, color, size, r, r0, r1, cp, g2, Direction.CLOCKWISE);
    }
    
	public void drawMarker( double value, Color color, float size, double r, double r0, double r1, Point2D cp, Graphics2D g2, Direction direction ) {
        if( direction == Direction.ANTICLOCKWISE ) {
        	value = maxValue - value + minValue;
        }
        double angleCurrent = angleStart + angleExtend * (1.0 - (value - minValue)/(maxValue - minValue));
        Point2D p1 = getPoint( cp, r * r0, angleCurrent );
        Point2D p2 = getPoint( cp, r * r1, angleCurrent );
        g2.setStroke( new BasicStroke(size));

        g2.setColor( color );
        g2.drawLine((int)p1.getX(),(int)p1.getY(),(int)p2.getX(),(int)Math.round(p2.getY()));		
	}
	
	public void drawArc( double start, double end, Color color, double r, double r0, double r1, Point2D cp, Graphics2D g2 ) {
		drawArc( start, end, color, r, r0, r1, cp, g2, Direction.CLOCKWISE );
	}
	public void drawArc( double start, double end, Color color, double r, double r0, double r1, Point2D cp, Graphics2D g2, Direction direction ) {
        if( direction == Direction.ANTICLOCKWISE ) {
        	start = maxValue - start + minValue;
        	end = maxValue - end + minValue;
        }
		Angle angleS = new Angle( angleStart + angleExtend * (1.0 - (start - minValue)/(maxValue - minValue)));
        double angleE = -1 * ( angleExtend * (end - start) / (maxValue - minValue));
        r0 *= r;
        r1 *= r;

        drawMarker( 180, color, 2, r, r0, r1, cp, g2 );
        drawMarker( start, color, 2, r, r0, r1, cp, g2 );
        drawMarker( end, color, 2, r, r0, r1, cp, g2 );
        //System.out.println("drawArc: "+ angleS +", "+angleE +": "+ start +", "+ end );
        
        // inner Arc
        Arc2D arc0 = new Arc2D.Double(Arc2D.PIE); 
        arc0.setFrameFromCenter( cp, new Point2D.Double(cp.getX() - r0,cp.getY() - r0) );
        arc0.setAngleStart( angleS.getDegrees() );
        arc0.setAngleExtent( angleE );

        // outer Arc
        Arc2D arc1 = new Arc2D.Double(Arc2D.PIE); 
        arc1.setFrameFromCenter( cp, new Point2D.Double(cp.getX() - r1,cp.getY() - r1) );
        arc1.setAngleStart( angleS.getDegrees() );
        arc1.setAngleExtent( angleE );

        Area area = new Area(arc1);
        area.subtract( new Area(arc0) );
        
		g2.setColor( color );
        g2.fill( area );
	}

    
    // the public bean methods
    
    /**
     * Gets the background color.
     * 
     * @return the background color
     */
    public Color getBackgroundColor() {
        return backgroundColor;
    }

    /**
     * Sets the background color.
     * 
     * @param bg the new background color
     */
    public void setBackgroundColor( Color bg ) {
        if( bg != null ) {
            this.backgroundColor = bg;
        }
    }

    
    /**
     * Gets the border background color.
     * 
     * @return the border background color
     */
    public Color getBackgroundBorderColor() {
        return backgroundBorderColor;
    }

    /**
     * Sets the border background color.
     * 
     * @param bg the new border background color
     */
    public void setBackgroundBorderColor( Color bg ) {
        if( bg != null ) {
            this.backgroundBorderColor = bg;
        }
    }

    
    /**
     * Gets the value.
     * 
     * @return the value
     */
    public double getValue() {
        return value;
    }

    /**
     * Sets the value.
     * 
     * @param value the new value
     */
    public void setValue( double value ) {
        this.value = value;
    }
    
        
    /**
     * Gets the min value.
     * 
     * @return the min value
     */
    public double getMinValue() {
        return minValue;
    }

    /**
     * Sets the min value.
     * 
     * @param minValue the new min value
     */
    public void setMinValue( double minValue ) {
        this.minValue = minValue;
    }
    

    /**
     * Gets the max value.
     * 
     * @return the max value
     */
    public double getMaxValue() {
        return maxValue;
    }

    /**
     * Sets the max value.
     * 
     * @param maxValue the new max value
     */
    public void setMaxValue( double maxValue ) {
        this.maxValue = maxValue;
    }
    

    /**
     * Gets the minor tick step.
     * 
     * @return the minor tick step
     */
    public double getMinorTickStep() {
        return minorTickStep;
    }

    /**
     * Sets the minor tick step.
     * 
     * @param minorTickStep the new minor tick step
     */
    public void setMinorTickStep( double minorTickStep ) {
        this.minorTickStep = minorTickStep;
    }

    /**
     * Gets the major tick step.
     * 
     * @return the major tick step
     */
    public double getMajorTickStep() {
        return majorTickStep;
    }

    /**
     * Sets the major tick step.
     * 
     * @param majorTickStep the new major tick step
     */
    public void setMajorTickStep( double majorTickStep ) {
        this.majorTickStep = majorTickStep;
    }
    
    /**
     * Gets the decimal place count.
     * 
     * @return the decimal place count. the range is [0..3]
     */
    public int getDecPlaceCount() {
        return decPlaceCount;
    }

    /**
     * Sets the decimal place count.
     * @param decPlaceCount the new decimal place count. the range is [0..3].
     */
    public void setDecPlaceCount( int decPlaceCount ) {
        if( this.decPlaceCount != decPlaceCount ) {
            this.decPlaceCount = Math.max( 0, Math.min( decPlaceCount, 3 ) ); // [0..3] - min 0 and max 3
            switch( this.decPlaceCount ) {
                case 0:
                    format = NumberFormat.getIntegerInstance();
                    break;
                case 1:
                    format = new DecimalFormat( "#0.0" );
                    break;
                case 2:
                    format = new DecimalFormat( "#0.00" );
                    break;
                default:
                    format = new DecimalFormat( "#0.000" );
            }
        }
    }
    
    
    /**
     * Gets the text.
     * 
     * @return the text
     */
    public String getText() {
        return text;
    }

    /**
     * Sets the text.
     * 
     * @param text the new text
     */
    public void setText( String text ) {
        this.text = text;
    }

    
    
    /**
     * Gets the scale line color.
     * 
     * @return the scale line color
     */
    public Color getScaleLineColor() {
        return scaleLineColor;
    }

    /**
     * Sets the scale line color.
     * 
     * @param scaleLineColor the new scale line color
     */
    public void setScaleLineColor( Color scaleLineColor ) {
        this.scaleLineColor = scaleLineColor;
    }

    /**
     * Gets the needle circle color.
     * 
     * @return the needle circle color
     */
    public Color getNeedleCircleColor() {
        return needleCircleColor;
    }

    /**
     * Sets the needle circle color.
     * 
     * @param needleCircleColor the new needle circle color
     */
    public void setNeedleCircleColor( Color needleCircleColor ) {
        this.needleCircleColor = needleCircleColor;
    }

    /**
     * Gets the needle color.
     * 
     * @return the needle color
     */
    public Color getNeedleColor() {
        return needleColor;
    }

    /**
     * Sets the needle color.
     * 
     * @param needleColor the new needle color
     */
    public void setNeedleColor( Color needleColor ) {
        this.needleColor = needleColor;
    }
    
 
    /**
     * Checks if is show value.
     * 
     * @return true, if is show value
     */
    public boolean isShowValue() {
        return showValue;
    }

    /**
     * Sets the show value.
     * 
     * @param showValue the new show value
     */
    public void setShowValue( boolean showValue ) {
        this.showValue = showValue;
    }
    
    
    /**
     * Gets the major tick color.
     * 
     * @return the major tick color
     */
    public Color getMajorTickColor() {
        return majorTickColor;
    }

    /**
     * Sets the major tick color.
     * 
     * @param majorTickColor the new major tick color
     */
    public void setMajorTickColor( Color majorTickColor ) {
        this.majorTickColor = majorTickColor;
    }

    /**
     * Gets the text color.
     * 
     * @return the text color
     */
    public Color getTextColor() {
        return textColor;
    }
    
    /**
     * Sets the text color.
     * 
     * @param textColor the new text color
     */
    public void setTextColor( Color textColor ) {
        this.textColor = textColor;
    }

    /**
     * Gets the angle extend.
     * 
     * @return the angle extend.  the range is [90..270]
     */
    public double getAngleExtend() {
        return angleExtend;
    }
    
    /**
     * Sets the angle extend.
     * 
     * @param angleExtend the new angle extend. the range is [90..270]
     */
    public void setAngleExtend( double angleExtend ) {
        this.angleExtend = angleExtend;

    }
    
    /**
     * Gets the angle start.
     * 
     * @return the angle start. the range is [0..360]
     */
    public double getAngleStart() {
        return angleStart;
    }
    
    /**
     * Sets the angle start.
     * 
     * @param angleStart the new angle start. the range is [-90..90]
     */
    public void setAngleStart( double angleStart ) {
        this.angleStart = angleStart;
    }

	public void setTextPosition(double x, double y) {
		this.textPosition = new Point2D.Double(x, y);
	}

    public void setTextPosition(Point2D textPosition) {
		this.textPosition = textPosition;
	}

	public Point2D getTextPosition() {
		return textPosition;
	}

	public void setTextHeight(double textHeight) {
		this.textHeight = textHeight;
	}

	public double getTextHeight() {
		return textHeight;
	}
}
