package org.gaugebook;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.geom.Point2D;
import java.awt.image.VolatileImage;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Vector;

import javax.swing.JPanel;



/**
 * class AbstractGauge
 * 
 * Base class of all gauges - no matter how they are displayed. In most cases it's not the 
 * parent class of a gauge, since its child classes AbstractClockGauge and AbstractGlassGauge often 
 * do a lot of work for you. So derive from those ;)
 */
public abstract class AbstractGauge extends JPanel implements KeyListener {

	private static final long serialVersionUID = 1L;
	
	protected GaugeBook owner = null;
	protected String title = null;

	protected double angleExtend = 270;   // [90..270]
    protected double angleStart = -45;    // [-90..90]
    protected double value = 0.0;
    protected double minValue = 0.0;
    protected double maxValue = 100.0;
    protected double minorTickStep = 2.5;
    protected double majorTickStep = 10.0;

	
	protected double border_width = .05;
	protected boolean antialias;
	protected boolean initdone = false;
	protected boolean skipFrame = false;
	
    private Vector<Scale> scales = new Vector<Scale>();
	private HashMap<String,Object> properties = new HashMap<String,Object>(20);

	protected VolatileImage vi = null;
	protected VolatileImage backVi = null;

	private Thread thread;
	
	// fkey management is more alpha than beta - to be improved...
	private String[] fkeylabels = null;
	private String[] fkeyevents = null;
	private boolean[] fkeyleds  = null;
	public String[] getFkeylabels() { return fkeylabels; }
	public String[] getFkeyevents() { return fkeyevents; }
	public boolean[] getFkeyleds() { return fkeyleds; }


	public AbstractGauge() {
		owner = null;
		title = null;
	}

	public void init( GaugeBook owner, String title ) {
		this.owner = owner;
		this.title = title;
		this.antialias = true;
		setBackground( Color.DARK_GRAY );
	    initdone = true;
	}

	public void update() {
		// update values here
		// and decide wether to call repaint() or not
	}

    public void setAntialias( boolean antialias ) {
    	this.antialias = antialias;
    }

    public static Point2D getPoint(Point2D cp, float r, float alpha){
        float angle = (float) Math.toRadians( alpha );
        float x = (float) (cp.getX() + r * Math.cos( angle ));
        float y = (float) (cp.getY() - r * Math.sin( angle ));
        return new Point2D.Float(x, y);
    }

    /**
     * Gets the string size.
     * 
     * @param s the string
     * @param g the graphics object
     * 
     * @return the string size
     */
    protected Dimension getStringSize( String s, Graphics g ) {
        FontMetrics fm = g.getFontMetrics();
        int width = fm.stringWidth( s );
		int height = fm.getMaxAscent() - fm.getMaxDescent();
        return new Dimension( width, height );
    }

    
    /**
     * Returns the Point from center point bei alpha Angle.
     * 
     * @param cp - the center point.
     * @param r the Radius
     * @param alpha - [-90,360]
     * 
     * @return the Point from center point bei Angle
     */
    protected Point2D getPoint(Point2D cp, double r, double alpha){
        double angle = Math.toRadians( alpha );
        double x = cp.getX() + r * Math.cos( angle );
        double y = cp.getY() - r * Math.sin( angle );
        return new Point2D.Double(x, y);
    }

    /**
     * Draw the specified text, aligned at the center of cp.
     * 
     * @param r the Radius of Gauge
     * @param cp the center Point of Gauge
     * @param g2 the graphics device
     * @param position alter position of text related to cp
     * @param textHeight the height of the Text (eg. 1 = normal, 1.2 = bigger, 0.8 = smaller)
     */
    protected void drawText( final double r, final Point2D cp, Graphics2D g2, Point2D position, double textHeight, String text ) {
    	if( text == null || text.length() == 0 ) return;
    	
        double r0 = r * .9;  
        double r1 = r; 
        double h = ( r1 - r0 ) * textHeight;   // height of text 

        g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h ) ) );
        
        Dimension dim = getStringSize( text, g2 );
        double x0 = cp.getX() - dim.getWidth()/2;
        double y0 = cp.getY() - dim.getHeight()/2;

        x0 = x0 + r * position.getX();
        y0 = y0 + r * position.getY();
        
        g2.setColor( Color.LIGHT_GRAY );
       	g2.drawString( text,(float)x0,(float)y0);
    }

    protected void drawText( final double r, final Point2D cp, Graphics2D g2, double textHeight, String text ) {
    	Point2D p0 = new Point2D.Double(0,0);
    	drawText( r, cp, g2, p0, textHeight, text);
    }

    protected void drawTextRight( final double r, final Point2D cp, Graphics2D g2, double textHeight, String text ) {
    	if( text == null || text.length() == 0 ) return;
    	
        double r0 = r * .9;  
        double r1 = r; 
        double h = ( r1 - r0 ) * textHeight;   // height of text 

        g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h ) ) );
        
        Dimension dim = getStringSize( text, g2 );
        double x0 = cp.getX() - dim.getWidth();
        double y0 = cp.getY() - dim.getHeight()/2;

       	g2.drawString( text,(float)x0,(float)y0);
    }

    protected void drawTextLeft( final double r, final Point2D cp, Graphics2D g2, double textHeight, String text ) {
    	if( text == null || text.length() == 0 ) return;
    	
        double r0 = proz( r, 90 );  
        double r1 = proz( r, 100 ); 
        double h = ( r1 - r0 ) * textHeight;   // height of text 

        g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h ) ) );
        
        Dimension dim = getStringSize( text, g2 );
        double x0 = cp.getX() ;
        double y0 = cp.getY() - dim.getHeight()/2;

       	g2.drawString( text,(float)x0,(float)y0);
    }


	// leading zero
	public String lz( int i, int len ) {
		String result = ""+ i;
		while(result.length() < len) {
			result = "0"+ i;
		}
		return result;
	}

	// leading zero
	public String lz( float f, int len ) {
		String result = ""+ f;
		while(result.length() < len) {
			result = "0"+ f;
		}
		return result;
	}

    protected void drawLine( Point2D p1, Point2D p2, Graphics2D g2 ) {
    	g2.drawLine( (int)Math.round( p1.getX() ), (int)Math.round( p1.getY() ),
    				 (int)Math.round( p2.getX() ), (int)Math.round( p2.getY() ));
    }
    
    /**
     * Sets a property of this gauge.
     * 
     * @param property
     * @param value
     */
    public void setProperty(String property, Object value) {
		String strval = value.toString();
		if( property.equals("minValue") || property.equals("maxValue") || property.equals("minorTickStep") || property.equals("majorTickStep")
			|| property.equals("scaleFactor") || property.equals("valueFactor") || property.equals("valueAddend")) {
			properties.put( property, Double.parseDouble(strval));
		}
		else if( property.equals("minValueR") || property.equals("maxValueR") || property.equals("minorTickStepR") || property.equals("majorTickStepR")
				|| property.equals("scaleFactorR") || property.equals("valueFactorR") || property.equals("valueAddendR")) {
				properties.put( property, Double.parseDouble(strval));
			}
		else if( property.equals("angleStart") || property.equals("angleExtend")) {
			properties.put( property, Integer.parseInt(strval));
		}
		else { 
			properties.put( property, strval );
		}
	}
	
    /**
     * Determines if this gauge has the given property.
     * 
     * @param property
     * @return true if the property exists, else false
     */
	public boolean hasProperty(String property) {
		return properties.containsKey(property);
	}
	
	/**
	 * Updates the gauge to use the latest given properties.
	 */
	public void updateProperties() {
		try {
			minValue = (Double)getProperty("minValue");
			maxValue = (Double)getProperty("maxValue");
			minorTickStep = (Double)getProperty("minorTickStep");
			majorTickStep = (Double)getProperty("majorTickStep");
			angleStart = (Integer)getProperty("angleStart");
			angleExtend = (Integer)getProperty("angleExtend");
			if(owner.debug)
				for( Object o: properties.keySet() ) {
					System.out.println("Properties: "+ o +": "+ properties.get(o));
				}
		} catch( Exception e ) {}
	}

	/**
	 * Get the value of a property of this gauge
	 * @param property
	 * @return the value of the given property
	 */
	public Object getProperty(String property) {
		return properties.get(property);
	}

	// KeyListener methods
	@Override
	public void keyPressed(KeyEvent e) {}

	@Override
	public void keyReleased(KeyEvent arg0) {}

	@Override
	public void keyTyped(KeyEvent arg0) {}

    // the following methods will be removed in the final:
    protected double proz( double v, double proz ) {
        return v * proz / 100.0;
    }

    private NumberFormat format = NumberFormat.getIntegerInstance();
    protected int decPlaceCount = 0; //[0..3]

    protected String valueToString(double v){
        return format.format( v );
    }
}
