package org.gaugebook;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Vector;

import org.gaugebook.math.Trigonometry;

public class Scale 
{
	Point2D cp;				// center point of scale
	float scaleR;			// scale radius
	int angleStart;			// scale starts at this angle
	int angleExtend;		// scale extends to this angle
	float minValue;			// minimum value
	float maxValue;			// maximum value
	float minorTickStep; 	// minor tick step
	float majorTickStep; 	// major tick step
	float minorTickStart; 	// minor ticks start at this angle
	float minorTickExtend; 	// minor ticks extend to this angle
	float majorTickStart;	// major ticks start at this angle
	float majorTickExtend;	// major ticks extend to this angle
	Color minorTickColor;
	Color majorTickColor;
	int decPlaceCount; 
	NumberFormat format;
	
	protected enum Direction { CLOCKWISE, ANTICLOCKWISE };
	
	private Direction direction = Direction.CLOCKWISE;
	
	Vector<ScaleNeedle> needles = new Vector<ScaleNeedle>();
	Vector<ScaleMarker> markers = new Vector<ScaleMarker>();
	Vector<ScaleArc> 	arcs	= new Vector<ScaleArc>();
	
	AbstractGauge owner;
	
	public Scale( AbstractGauge owner, Point2D cp, float scaleR, Direction direction ) {
		this.owner = owner;
		this.cp = cp;
		this.scaleR = scaleR;
		this.direction = direction;
	}
	
	public void draw( double r, Graphics2D g2 ) {
		// draw arcs
		for( int i=0; i < arcs.size(); i++ ) {
			arcs.get(i).draw(g2);
		}
		// draw markers
		for( int i=0; i < markers.size(); i++ ) {
			markers.get(i).draw(g2);
		}
		// draw scale
        double r0 = r * .65; // 55
        double r1 = r * .8; // 58
        
        double majorTickSize = r * .7;
        double minorTickSize = r * .5;
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
            g2.setColor( minorTickColor );
            g2.setStroke( new BasicStroke(1.f) );            
            double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            for(double i= angleStart + angleExtend; i >= angleStart ; i-=minorGradSteps){
                Point2D p0 = Trigonometry.getPoint( cp, minorTickR, i );
                Point2D p1 = Trigonometry.getPoint( cp, minorTickR+minorTickSize, i );
                g2.draw( new Line2D.Double(p0,p1) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0, r * .12 * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
            	double myv = direction == Direction.CLOCKWISE ? v : maxValue - v + minValue; 
            			
                double currentAngle = angleStart + angleExtend - stepAngle * (v - minValue) / majorTickStep;
                Point2D p0 = Trigonometry.getPoint( cp, majorTickR, currentAngle );
                Point2D p1 = Trigonometry.getPoint( cp, majorTickR + majorTickSize, currentAngle );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s = valueToString( myv );
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = Trigonometry.getPoint( cp, fontR - delta, currentAngle ); // marker center point
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
                
            }
            
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }		
	}
	
	public void drawNeedles( Graphics2D g2 ) {
		for( int i=0; i < needles.size(); i++ ) {
			needles.get(i).draw(g2);
		}
	}

    /**
     * Gets the string size.
     * 
     * @param s the string
     * @param g the graphics object
     * 
     * @return the string size
     */
    protected Dimension getStringSize( String s, Graphics g ) {
        FontMetrics fm = g.getFontMetrics();
        int width = fm.stringWidth( s );
		int height = fm.getMaxAscent() - fm.getMaxDescent();
//        int height = (int)fm.getLineMetrics(s, g).getHeight();
/*
         Rectangle2D rect = fm.getStringBounds(s, g);
        int width = (int)Math.round(rect.getWidth());
        int height= (int)Math.round(rect.getHeight());
*/        
        return new Dimension( width, height );
    }
    
    /**
     * converts value to String.
     * 
     * @param v the value that is to converted.
     * 
     * @return the string representation of value
     */
    protected String valueToString(double v){
        return format.format( v );
    }
   
    /**
     * Gets the decimal place count.
     * 
     * @return the decimal place count. the range is [0..3]
     */
    public int getDecPlaceCount() {
        return decPlaceCount;
    }

    /**
     * Sets the decimal place count.
     * @param decPlaceCount the new decimal place count. the range is [0..3].
     */
    public void setDecPlaceCount( int decPlaceCount ) {
        if( this.decPlaceCount != decPlaceCount ) {
            this.decPlaceCount = Math.max( 0, Math.min( decPlaceCount, 3 ) ); // [0..3] - min 0 and max 3
            switch( this.decPlaceCount ) {
                case 0:
                    format = NumberFormat.getIntegerInstance();
                    break;
                case 1:
                    format = new DecimalFormat( "#0.0" );
                    break;
                case 2:
                    format = new DecimalFormat( "#0.00" );
                    break;
                default:
                    format = new DecimalFormat( "#0.000" );
            }
        }
    }
}