/* G1000 gauge for GaugeBook
 * 
 * Copyright (C) 2009 Michael Nagler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.gaugebook.gauges.aviation.glass;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Event;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfigTemplate;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.Transparency;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.ColorModel;
import java.awt.image.VolatileImage;
import java.text.NumberFormat;
import java.util.Locale;

import org.gaugebook.AbstractGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;

/**
 * G1000 Gauge
 * 
 * This class imitates the Garmin G1000(R) display for flight simulation. 
 * G1000 is remarked by Garmin Ltd, Kansas, USA (http://www.garmin.com).
 * 
 * <h3>License</h3>
 * This is free, open source software, license: LGPL.
 * 
 * @author Michael Nagler
 */
public class G1000 extends AbstractGauge {

    // Colors
	private static final Color skyColorAI = new Color(66,99,153);
	private static final Color gndColorAI = new Color(74,68,43);
	
	private static final Color clrBlack = new Color(0,0,0);
	private static final Color clrWhite = new Color( 230, 230, 230 );
	private static final Color clrGrey  = Color.LIGHT_GRAY;
	private static final Color clrGreen = new Color( 108, 168, 107 );
	private static final Color clrPink  = new Color( 168, 108, 167 );
	private static final Color clrRed   = new Color( 222, 0, 0 );
	private static final Color clrBlue  = new Color( 121, 135, 158 );
    
    // FSX Variables
	private FSXVariable varMasterSwitchOn;
	private FSXVariable varCom1Freq;
	private FSXVariable varCom2Freq;
	private FSXVariable varCom1Stby;
	private FSXVariable varCom2Stby;
	private FSXVariable varNav1Freq;
	private FSXVariable varNav2Freq;
	private FSXVariable varNav1Stby;
	private FSXVariable varNav2Stby;
	private FSXVariable varXpdrCode; 

	private FSXVariable com1Xmit, com2Xmit, comBoth;
	
	private FSXVariable indicatedAirspeed;
	private FSXVariable trueAirspeed;
	private FSXVariable selectedAirspeed;
	private FSXVariable selectedMach;
	private FSXVariable indicatedAltitude;
	private FSXVariable selectedAltitude;
	private FSXVariable kohlsmanMB;
	private FSXVariable kohlsmanHG;
	private FSXVariable varAviCircuitOn, varAviMasterOn;

	private FSXVariable verticalSpeed;
	private FSXVariable bankVar;
	private FSXVariable pitchVar;
	private FSXVariable ballVar;
	private FSXVariable headingIndicator, headingLockDir;
	private FSXVariable vorOBS, vorCDI, vorToFrom, vorGSAngle, vorGSFlag;
	
	private FSXVariable nav1ident, nav1name;
	private FSXVariable nav2ident, nav2name;
	private FSXVariable nav1obs, nav2obs;
	private FSXVariable navHasNav1, navHasNav2, hasDME1, hasDME2, navDME1, navDME2;
	private FSXVariable hsiHasLoc, hsiBearing, hsiBearingValid, hsiCDIValid, hsiCDI, hsiID;
	
    private FSXVariable vor2Radial;
    private FSXVariable adfFreq, adfRadial, adfSignal, adfIdent;

    private FSXVariable gpsGroundTrueTrack;
    private FSXVariable gpsGroundTrueHeading;
    private FSXVariable gpsGroundMagneticTrack;
    private FSXVariable gpsWPDistance;
    private FSXVariable gpsWPDesiredTrack;
    private FSXVariable gpsWPNextID;
    private FSXVariable gpsApproachType, gpsDrivesNav1;
    
    private FSXVariable flightDirectorActive;
	private FSXVariable flightDirectorPitch;
	private FSXVariable flightDirectorBank;
    
	private FSXVariable varLocalTime;
	private FSXVariable varOAT;
	
	private FSXVariable apHdgLock, apSpdLock, apAltLock;
	
	private FSXVariable vs0speed, vs1speed, vcspeed, maxmach, asbarberpole;
	private FSXVariable markerI, markerM, markerO;

	private long headingChanged = 0;
    private long obsChanged = 0;
    private int editNav = 1;
    private int editCom = 1;
	
    private Stroke lineStroke; 

	private double 	bank=0, 	// bank
					pitch=0,	// pitch
					fdbank=0,	// flight director bank
					fdpitch=0,	// flight director pitch
					ias,		// indicated airspeed
					sas,		// selected airspeed
					tas, 		// true airspeed
					vs,			// vertical speed
					smach, 		// selected mach
					hi,			// heading indicated
					hld,		// heading lock dir
					qnh, 		// kohlsman setting mb
					ialt, 		// indicated altitude
					salt,		// selected altitude
					shdg,		// selected heading
					hdg,		// magnetic heading
					obs, cdi, tf, gsa, gsf, vor2r, adfr, com1s, com2s, nav1s, nav2s;
					;

	double gpsgth, gpswd, gpswdt, gpsgmt, gpsgtt; String gpswni;

	private boolean aviOn = false, aviOld = false, flightDirOn = false, hdglock = false, spdlock = false, altlock = false;
	
	private int lastWidth = 0;
	private int lastHeight = 0;
	private boolean updatebg = true;
	private boolean lastAviOn = false;

	private static NumberFormat dmeNF  = NumberFormat.getInstance( Locale.US );
	private static NumberFormat freqNF = NumberFormat.getInstance( Locale.US );
	private static NumberFormat distNF = NumberFormat.getInstance( Locale.US );

	private static final long serialVersionUID = 1L;
	
	private long framecounter = 0;

	public void init( GaugeBook owner, String title ) { 
		super.init( owner, title );
		System.out.println( "gaugebook size: "+ owner.getWidth() +", "+ owner.getHeight());

		varMasterSwitchOn = owner.registerFSXVariable("ELECTRICAL MASTER BATTERY");
			
		varCom1Freq = owner.registerFSXVariable("COM ACTIVE FREQUENCY:1");
		varCom2Freq = owner.registerFSXVariable("COM ACTIVE FREQUENCY:2");
		varCom1Stby = owner.registerFSXVariable("COM STANDBY FREQUENCY:1");
		varCom2Stby = owner.registerFSXVariable("COM STANDBY FREQUENCY:2");
		varNav1Freq = owner.registerFSXVariable("NAV ACTIVE FREQUENCY:1");
		varNav2Freq = owner.registerFSXVariable("NAV ACTIVE FREQUENCY:2");
		varNav1Stby = owner.registerFSXVariable("NAV STANDBY FREQUENCY:1");
		varNav2Stby = owner.registerFSXVariable("NAV STANDBY FREQUENCY:2");
		
		com1Xmit = owner.registerFSXVariable("COM TRANSMIT:1");
		com2Xmit = owner.registerFSXVariable("COM TRANSMIT:2");
		comBoth  = owner.registerFSXVariable("COM RECIEVE ALL");

        indicatedAirspeed = owner.registerFSXVariable("AIRSPEED INDICATED"); 
        selectedAirspeed = owner.registerFSXVariable("AUTOPILOT AIRSPEED HOLD VAR");
        selectedMach = owner.registerFSXVariable("AUTOPILOT MACH HOLD VAR");
        trueAirspeed = owner.registerFSXVariable("AIRSPEED TRUE"); 
	    indicatedAltitude = owner.registerFSXVariable("INDICATED ALTITUDE");
	    selectedAltitude = owner.registerFSXVariable("AUTOPILOT ALTITUDE LOCK VAR");
	    kohlsmanMB = owner.registerFSXVariable("KOHLSMAN SETTING MB");
	    kohlsmanHG = owner.registerFSXVariable("KOHLSMAN SETTING HG");    
		verticalSpeed = owner.registerFSXVariable("VERTICAL SPEED");
	    bankVar = owner.registerFSXVariable("ATTITUDE INDICATOR BANK DEGREES");
	    pitchVar = owner.registerFSXVariable("ATTITUDE INDICATOR PITCH DEGREES");
	    ballVar = owner.registerFSXVariable("TURN COORDINATOR BALL");
        
	    headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");
        headingLockDir = owner.registerFSXVariable("AUTOPILOT HEADING LOCK DIR");
        vorOBS = owner.registerFSXVariable("NAV OBS:1");
        vorCDI = owner.registerFSXVariable("NAV CDI:1");
        vorToFrom = owner.registerFSXVariable("NAV TOFROM:1");
		vorGSAngle = owner.registerFSXVariable("NAV GLIDE SLOPE ERROR:1");
		vorGSFlag = owner.registerFSXVariable("NAV GS FLAG:1");
	    vor2Radial = owner.registerFSXVariable("NAV RADIAL:2");
	    adfRadial = owner.registerFSXVariable("ADF RADIAL:1");
	    varXpdrCode = owner.registerFSXVariable("TRANSPONDER CODE:1");
	    varLocalTime = owner.registerFSXVariable("LOCAL TIME");
	    varOAT = owner.registerFSXVariable("AMBIENT TEMPERATURE");
		gpsGroundTrueTrack = owner.registerFSXVariable("GPS GROUND TRUE TRACK");
		gpsGroundTrueHeading = owner.registerFSXVariable("GPS GROUND TRUE HEADING");
		gpsGroundMagneticTrack = owner.registerFSXVariable("GPS GROUND MAGNETIC TRACK");
		adfFreq = owner.registerFSXVariable("ADF ACTIVE FREQUENCY:1");
		adfSignal = owner.registerFSXVariable("ADF SIGNAL:1");
		adfIdent = owner.registerFSXVariable("ADF IDENT");
		
	    gpsWPDistance = owner.registerFSXVariable("GPS WP DISTANCE");
	    gpsWPDesiredTrack = owner.registerFSXVariable("GPS WP DESIRED TRACK");
	    gpsWPNextID = owner.registerFSXVariable("GPS WP NEXT ID");
	    gpsApproachType = owner.registerFSXVariable("GPS APPROACH APPROACH TYPE");
	    
    	varCom1Freq = owner.registerFSXVariable("COM ACTIVE FREQUENCY:1");
    	varCom2Freq = owner.registerFSXVariable("COM ACTIVE FREQUENCY:2");
    	varCom1Stby = owner.registerFSXVariable("COM STANDBY FREQUENCY:1");
    	varCom2Stby = owner.registerFSXVariable("COM STANDBY FREQUENCY:2");
		
    	varNav1Freq = owner.registerFSXVariable("NAV ACTIVE FREQUENCY:1");
    	varNav2Freq = owner.registerFSXVariable("NAV ACTIVE FREQUENCY:2");
    	varNav1Stby = owner.registerFSXVariable("NAV STANDBY FREQUENCY:1");
    	varNav2Stby = owner.registerFSXVariable("NAV STANDBY FREQUENCY:2");

    	navHasNav1 = owner.registerFSXVariable("NAV HAS NAV:1");
    	navHasNav2 = owner.registerFSXVariable("NAV HAS NAV:2");
    	nav1ident = owner.registerFSXVariable("NAV IDENT:1");
    	nav1name = owner.registerFSXVariable("NAV NAME:1");
    	nav2ident = owner.registerFSXVariable("NAV IDENT:2");
    	nav2name = owner.registerFSXVariable("NAV NAME:2");
		nav1obs = owner.registerFSXVariable("NAV OBS:1");
		nav2obs = owner.registerFSXVariable("NAV OBS:2");

		gpsDrivesNav1 = owner.registerFSXVariable("GPS DRIVES NAV1");
		hsiHasLoc = owner.registerFSXVariable("HSI HAS LOCALIZER");
		hsiBearing = owner.registerFSXVariable("HSI BEARING");
		hsiBearingValid = owner.registerFSXVariable("HSI BEARING VALID");
		hsiCDI  = owner.registerFSXVariable("HSI CDI NEEDLE");
		hsiCDIValid = owner.registerFSXVariable("HSI CDI NEEDLE VALID");
		hsiID = owner.registerFSXVariable("HSI STATION IDENT");
		
		hasDME1 = owner.registerFSXVariable("NAV HAS DME:1");
		hasDME2 = owner.registerFSXVariable("NAV HAS DME:2");
		navDME1 = owner.registerFSXVariable("NAV DME:1");
		navDME2 = owner.registerFSXVariable("NAV DME:2");
		
		flightDirectorActive = owner.registerFSXVariable("AUTOPILOT FLIGHT DIRECTOR ACTIVE");
		flightDirectorPitch  = owner.registerFSXVariable("AUTOPILOT FLIGHT DIRECTOR PITCH");
		flightDirectorBank   = owner.registerFSXVariable("AUTOPILOT FLIGHT DIRECTOR BANK");
		
		apSpdLock = owner.registerFSXVariable("AUTOPILOT AIRSPEED HOLD");
		apHdgLock = owner.registerFSXVariable("AUTOPILOT HEADING LOCK");
		apAltLock = owner.registerFSXVariable("AUTOPILOT ALTITUDE LOCK");

		vs0speed = owner.registerFSXVariable("DESIGN SPEED VS0");
		vs1speed = owner.registerFSXVariable("DESIGN SPEED VS1");
		vcspeed  = owner.registerFSXVariable("DESIGN SPEED VC");
		asbarberpole = owner.registerFSXVariable("AIRSPEED BARBER POLE");
		maxmach = owner.registerFSXVariable("MACH MAX OPERATE");

		varAviMasterOn = owner.registerFSXVariable("AVIONICS MASTER SWITCH");
		varAviCircuitOn = owner.registerFSXVariable("CIRCUIT AVIONICS ON");

		markerI = owner.registerFSXVariable("INNER MARKER");
		markerM = owner.registerFSXVariable("MIDDLE MARKER");
		markerO = owner.registerFSXVariable("OUTER MARKER");
		
		freqNF.setMaximumFractionDigits(2);
		freqNF.setMinimumFractionDigits(2);

		dmeNF.setMaximumFractionDigits(1);
		dmeNF.setMinimumFractionDigits(1);
		dmeNF.setMaximumIntegerDigits(2);
		dmeNF.setMinimumIntegerDigits(1);
		
		distNF.setMaximumFractionDigits(1);
		distNF.setMinimumFractionDigits(1);
	}

	public void update() {
		if( shdg != headingLockDir.getDoubleValue() ) {
			headingChanged = System.currentTimeMillis(); 
		}
		
		if( obs != vorOBS.getDoubleValue() ) {
			obsChanged = System.currentTimeMillis();
		}
		
		if( com1s != varCom1Stby.getDoubleValue() ) editCom = 1;
		else if( com2s != varCom2Stby.getDoubleValue() ) editCom = 2;

		if( nav1s != varNav1Stby.getDoubleValue() ) editNav = 1;
		else if( nav2s != varNav2Stby.getDoubleValue() ) editNav = 2;

		ialt = indicatedAltitude.getDoubleValue();
		ias = indicatedAirspeed.getDoubleValue();
		sas = selectedAirspeed.getDoubleValue();
		tas = trueAirspeed.getDoubleValue();
		vs = verticalSpeed.getDoubleValue() * 60;
		smach = selectedMach.getDoubleValue();
		hi = ( 180 / StrictMath.PI ) * headingIndicator.getDoubleValue();
		hld = headingLockDir.getDoubleValue();
		salt = selectedAltitude.getDoubleValue();
		pitch = ( 180 / StrictMath.PI ) * pitchVar.getDoubleValue();
		bank = ( 180 / StrictMath.PI ) * bankVar.getDoubleValue();
		fdpitch = ( 180 / StrictMath.PI ) * flightDirectorPitch.getDoubleValue();
		fdbank = ( 180 / StrictMath.PI ) * flightDirectorBank.getDoubleValue();
		qnh = kohlsmanMB.getDoubleValue();
		hdg = hi;
		shdg = headingLockDir.getDoubleValue();

		gpsgtt = 180 / StrictMath.PI * gpsGroundTrueTrack.getDoubleValue();
	    gpsgth = 180 / StrictMath.PI * gpsGroundTrueHeading.getDoubleValue();
	    gpsgmt = 180 / StrictMath.PI * gpsGroundMagneticTrack.getDoubleValue();
	    gpswd = gpsWPDistance.getDoubleValue() / 1852.216;
	    gpswdt = 180 / StrictMath.PI * gpsWPDesiredTrack.getDoubleValue();
	    gpswni = gpsWPNextID.getStringValue();
	    	
		obs = vorOBS.getDoubleValue();
		cdi = vorCDI.getDoubleValue();
		tf  = vorToFrom.getDoubleValue();
	 	gsa = vorGSAngle.getDoubleValue();
		gsf = vorGSFlag.getDoubleValue();

		adfr = adfRadial.getDoubleValue();
		vor2r = vor2Radial.getDoubleValue();
		
		com1s = varCom1Stby.getDoubleValue();
		com2s = varCom2Stby.getDoubleValue();
		nav1s = varNav1Stby.getDoubleValue();
		nav2s = varNav2Stby.getDoubleValue();

		aviOn = varMasterSwitchOn.getBooleanValue() && varAviMasterOn.getBooleanValue() && varAviCircuitOn.getBooleanValue();
		flightDirOn = flightDirectorActive.getBooleanValue();
		hdglock = apHdgLock.getBooleanValue();
		spdlock = apSpdLock.getBooleanValue();
		altlock = apAltLock.getBooleanValue();

		//render( getGraphics() );
		repaint();
	}

	public void paint( Graphics g ) {
        int w = getWidth();
        int h = getHeight();

        do {
	        GraphicsConfiguration gc = this.getGraphicsConfiguration();
	        if( gc == null ) return;
	        
	    	if( updatebg || lastAviOn != aviOn || lastWidth != w || lastHeight != h || vi == null ) {
	    		vi = gc.createCompatibleVolatileImage( w, h, Transparency.OPAQUE ); // gc.getColorModel().getTransparency() );
	    		lastWidth = w;
	    		lastHeight = h;
	    		lastAviOn = aviOn;
	    		updatebg = true;
	    	}
	
	    	Graphics2D g2 = vi.createGraphics();
	        g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));
	
	        double r =  Math.min( w, h ) / 2 - 5;            // radius
	        if( r < 0 ) r = 100;
	        
	        // rw + rh to support different aspect ratios
	        if( w < 800 ) w = 160;
	        if( h < 480 ) h = 480;
	        if( w < h * 1.6 ) w = (int)(h * 1.6);
	        double rw = w / 2;
	        double rh = h / 2;
	
	        Point2D cp = new Point2D.Double( w / 2, h / 2 ); 
			lineStroke = new BasicStroke((int)Math.round(r*0.0025));
	
	        if( aviOn ) {
	            Point2D aicp = getPoint( cp, r * .325, 90 );
	            Point2D hsicp = new Point2D.Double( cp.getX(), cp.getY() + r * .45 );
	
	    		if( updatebg ) drawBackground( rw, rh, cp, g2 );
	            drawSky( rw, rh, cp, aicp, g2 );
	            drawAI( rw, rh, aicp, g2 );
	            drawHSI( rw, rh, hsicp, g2 );
	            drawASI( rw, rh, aicp, g2 );
	            drawALT( rw, rh, aicp, g2 ); 
		        drawText( rw, rh, cp, aicp, hsicp, g2 );
	
	        }
	        else {
	        	g2.setColor( clrBlack );
	        	g2.fillRect( 0, 0, w, h );
	        }
	        g2.dispose();
        }
	    while( vi.contentsLost() );
        
		// if( owner.debug ) System.out.println("repaint (frame #"+ framecounter++ +")");

        g.drawImage( vi, 0, 0, this );

    }	

	/**
	 * Draws the background of the G1000, that is the ground color fill + upper and lower text rects + lines.
	 * 
	 * @param rw
	 * @param rh
	 * @param cp
	 * @param g2
	 */
	protected void drawBackground( final double rw, final double rh, final Point2D cp, Graphics2D g2 ) {
		Point2D p1, p2;
		int w = getWidth();
		int h = getHeight();

        do {
            GraphicsConfiguration gc = getGraphicsConfiguration();

        	if( backVi == null || updatebg || backVi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
        		backVi = gc.createCompatibleVolatileImage( w, h, Transparency.OPAQUE );
        	}
        	
        	Graphics2D bg2 = backVi.createGraphics();

            bg2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias 
            		? RenderingHints.VALUE_ANTIALIAS_ON 
            		: RenderingHints.VALUE_ANTIALIAS_OFF ));
	        

//            Graphics2D bg2 = g2;
            
	    	// some 3d bevel would be nice here
			p1 = new Point2D.Double( cp.getX() - rw, cp.getY() - rh );
			Rectangle2D.Double rr = new Rectangle2D.Double( p1.getX(), p1.getY(), rw * 2, rh * 2 );
			bg2.fill(rr);
			bg2.setClip(rr);
			
			// return if avionics circuit is off
			if( aviOn && initdone )
			{
		    	// text indicator background = grey rect top
				//bg2.setColor(clrGrey);
				//p1 = new Point2D.Double(cp.getX() - r * .55, cp.getY() - r * .9 );
				//bg2.fillRect((int)p1.getX(),(int)p1.getY(), (int)(r*.9), (int)(r*.16));
		
		    	// ai background = center
				bg2.setColor(gndColorAI);
				p1 = new Point2D.Double(cp.getX() - rw, cp.getY() - rh * .85 );
				rr = new Rectangle2D.Double(p1.getX(), p1.getY(), rw * 2, rh * 1.77 );
		    	bg2.fill(rr);
		    	
		    	// static text
		        bg2.setColor( clrWhite );
		    	p1 = new Point2D.Double(cp.getX() - rw * .95, cp.getY() - rh * .925 );
		        drawText( rh, p1, bg2, .4, "NAV1" );
				p1 = new Point2D.Double(cp.getX() - rw * .95, cp.getY() - rh * .852 );
		        drawText( rh, p1, bg2, .4, "NAV2" );
		
				p1 = new Point2D.Double(cp.getX() + rw * .96, cp.getY() - rh * .925 );
		        drawTextRight( rh, p1, bg2, .4, "COM1" );
				p1 = new Point2D.Double(cp.getX() + rw * .96, cp.getY() - rh * .852 );
		        drawTextRight( rh, p1, bg2, .4, "COM2" );
		
		    	// upper bar lines
		        bg2.setColor( clrWhite );
		    	p1 = new Point2D.Double(cp.getX() - rw, cp.getY() - rh * .85 );
		    	p2 = new Point2D.Double(cp.getX() + rw, cp.getY() - rh * .85 );
		    	bg2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );
		
		    	p1 = new Point2D.Double(cp.getX() - rw * .6, cp.getY() - rh * 1 );
		    	p2 = new Point2D.Double(cp.getX() - rw * .6, cp.getY() - rh * .85 );
		    	bg2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );
		
		    	p1 = new Point2D.Double(cp.getX() + rw * .6, cp.getY() - rh * 1 );
		    	p2 = new Point2D.Double(cp.getX() + rw * .6, cp.getY() - rh * .85 );
		    	bg2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );
		
		    	p1 = new Point2D.Double(cp.getX() - rw * .6 , cp.getY() - rh * .925 );
		    	p2 = new Point2D.Double(cp.getX() + rw * .6, cp.getY() - rh * .925 );
		    	bg2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );
		    	
		    	// 
		        bg2.setColor( new Color( 200, 204, 208 ));
		    	String[] txt = {"WPT", "DIS", "DTK", "TRK"};
		        double delta = rw * 1.2 / 4; 
		        for( int i = 0; i < 4; i++ ) {
		        	p1 = new Point2D.Double(cp.getX() - rw * .58 + i * delta, cp.getY() - rh * .925 );
		        	drawTextLeft( rh, p1, bg2, .45, txt[i] );
		        }
		
		        g2.setColor( new Color( 150, 150, 150 ));
		
        	}
	        bg2.dispose();
        } while( backVi.contentsLost() );
    	g2.drawImage( backVi, 0, 0, this );
    }


	/**
	 * Draw Attitude Indicator
	 * 
	 * @param rw
	 * @param rh
	 * @param aicp
	 * @param g2
	 */
	protected void drawAI( double rw, final double rh, final Point2D aicp, Graphics2D g2 ) {
		Point2D p1, p2, p3, p4, p5, p6, p7;
    	Polygon pol;
    	Ellipse2D ell;
    	Rectangle2D rect;
    	Shape oldClip = g2.getClip();
        Font oldFont = g2.getFont();

		double pf = 45;
		Angle a = new Angle( -1 * bank );
		rw = rh;
		// clip
		// ----
 //       p1 = new Point2D.Double(cp.getX() - rw, cp.getY() - rh * 0.77 );
//		Rectangle2D rr = new Rectangle2D.Double(p1.getX(), p1.getY(), rw * 2, rh * 1.85 );
//		g2.setClip(rr);

        // pitch scale horizontal lines
        // ----------------------------
        ell = new Ellipse2D.Double( aicp.getX() - rw*.35, aicp.getY()-rh*.35, rw*.7, rh*.7);
        g2.setClip(ell);
		g2.setColor( clrWhite );
		g2.setStroke( lineStroke );
        double majorFontSize = Math.max( 6.0, rh * .05);  // minimum 6.0
        Font font = oldFont.deriveFont( Font.PLAIN, (float)majorFontSize );
        g2.setFont( font);
        Angle currentAngle = new Angle( -1 * bank );
        TextLayout stl;
        Shape ts;
    	AffineTransform tat = new AffineTransform();

		for( int i = -80; i <= 80; i += 10 ) {
			if( i == 0 ) continue;
			p1 = getPoint( aicp, rw * (pitch + i ) / pf, a.getDegreesPlus(90) );
			p1 = getPoint( p1, rw * .15, a.getDegrees() );
			p2 = getPoint( p1, rw * .3, a.getDegreesPlus(180));
			g2.draw( new Line2D.Double(p1, p2));
        	String s = valueToString(i ); //  Math.abs( i ));

        	currentAngle = new Angle( -1 * bank );
            stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(rh * 0.05)), new FontRenderContext(null, true, false));
        	tat = new AffineTransform();
        	currentAngle = new Angle( 360 - currentAngle.getDegrees() );
        	rect = stl.getBounds();
        	tat.translate(aicp.getX() - rect.getWidth()/2, aicp.getY() + stl.getAscent()/2.5 ); 
        	tat.rotate( Math.toRadians( currentAngle.getDegrees() ), rect.getCenterX(), rect.getCenterX() );
        	tat.translate( rw *.2, -rh * ( pitch +i ) / pf );
        	ts = stl.getOutline(tat);

        	g2.fill(ts);
        	tat.translate( -rw *.4, 0 );
        	ts = stl.getOutline(tat);
            g2.fill(ts);
    	}
		for( int i = -35; i < 40; i += 10 ) {
			p1 = getPoint( aicp, rw * (pitch + i ) / pf, a.getDegreesPlus(90) );
			p1 = getPoint( p1, rw * .09, a.getDegrees() );
			p2 = getPoint( p1, rw * .18, a.getDegreesPlus(180));
			g2.draw( new Line2D.Double(p1, p2)); 
    	}
		for( float i = -17.5f; i < 20; i += 5 ) {
			p1 = getPoint( aicp, rw * (pitch + i ) / pf, a.getDegreesPlus(90) );
			p1 = getPoint( p1, rw * .04, a.getDegrees() );
			p2 = getPoint( p1, rw * .08, a.getDegreesPlus(180));
			g2.draw( new Line2D.Double(p1, p2)); 
    	}
    	g2.setClip(null);

    	// todo: red pitch warning markers
    	
		// upper triangle
		a = new Angle( 0 );
        p1 = getPoint( aicp, rw * .41, a.getDegreesPlus(90));
        p2 = getPoint( p1, rw * .055, a.getDegreesPlus(60));
        p3 = getPoint (p1, rw * .055, a.getDegreesPlus(120));
        
        pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );

        g2.setColor( clrWhite );
        g2.fill( pol);

        // lower triangle
		a = new Angle( -bank );
        p1 = getPoint( aicp, rw * .39, a.getDegreesPlus(90));
        p2 = getPoint( p1, rw * .055, a.getDegreesPlus(240));
        p3 = getPoint (p1, rw * .055, a.getDegreesPlus(300));
        
        pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );

		g2.setStroke( lineStroke );
        //g2.setColor( skyColorAI );
        //g2.fill( pol);
        g2.setColor( clrWhite );
        g2.fill( pol);
        
        // turn coordinator
        double ball = ballVar.getDoubleValue();
        
        Angle angleCurrent = new Angle( -bank +90 );
		p1 = getPoint( aicp, rw*.325, angleCurrent.getDegrees() );
		p2 = getPoint( p1, -ball / 1024 * rw, angleCurrent.plus(90));
		p3 = getPoint( p2, rw * .03 , angleCurrent.plus(90));
		p4 = getPoint( p2, rw * .03 , angleCurrent.plus(270));
		g2.setStroke(new BasicStroke((int)Math.round(rw*0.01)));
		g2.draw( new Line2D.Double( p3, p4 ));


        // top scale lines
        g2.setColor( clrWhite );
        Point2D ap, np;
		g2.setStroke( lineStroke );
        double f = .75;
        for( int a1 = -60; a1 <= 60; a1 += 10 ) {
        	if( a1 == 0 || Math.abs(a1)==80 || Math.abs(a1)==70 || Math.abs(a1)==40 || Math.abs(a1)==50 ) continue;
        	f = Math.abs(a1) == 10 || Math.abs(a1) == 20 ? .44 : .46;
            angleCurrent = new Angle( a1 + 90 ); // new Angle( bank + 90 + -1 * bank + a );
            ap = getPoint( aicp, rw * .41, angleCurrent.getDegrees() );
            np = getPoint( aicp, rw * f, angleCurrent.getDegrees() );
            g2.draw( new Line2D.Double(ap, np));
        }
		g2.setClip(oldClip);
		// 45 deg lines
        ap = getPoint( aicp, rw * .41, 135 );
        np = getPoint( aicp, rw * .44, 135 );
        g2.draw( new Line2D.Double(ap, np));

        ap = getPoint( aicp, rw * .41, 45 );
        np = getPoint( aicp, rw * .44, 45 );
        g2.draw( new Line2D.Double(ap, np));
        
        // top scale combining arc
        Arc2D arc = new Arc2D.Double( new Rectangle2D.Double( aicp.getX() - rw * .41, aicp.getY() - rh * .41, rw * .82, rh * .82 ), 30, 120, Arc2D.OPEN ); 
        g2.draw( arc );
        
        // bank+pitch 0 indicator rects + shapes
        rect = new Rectangle2D.Double(aicp.getX() - rw * .015, aicp.getY() - rh * .015, rw * .03, rh * .03 );
        g2.setColor(clrBlack);
        g2.fill(rect);
        g2.setColor(clrWhite);
        g2.draw(rect);
        // left shape
		p1 = new Point2D.Double(aicp.getX() - rw * .16, aicp.getY() - rh * .015 );
		p2 = new Point2D.Double(aicp.getX() - rw * .36, aicp.getY() - rh * .015 );
		p3 = new Point2D.Double(aicp.getX() - rw * .36, aicp.getY() + rh * .01 );
		p4 = new Point2D.Double(aicp.getX() - rw * .19, aicp.getY() + rh * .01 );
		p5 = new Point2D.Double(aicp.getX() - rw * .19, aicp.getY() + rh * .030 );
		p6 = new Point2D.Double(aicp.getX() - rw * .16, aicp.getY() + rh * .030 );
		p7 = new Point2D.Double(aicp.getX() - rw * .16, aicp.getY() - rh * .015 );
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
		g2.setColor(clrBlack);
		g2.fill(pol);
		g2.setColor(clrWhite);
		g2.draw(pol);
		// right shape
		p1 = new Point2D.Double(aicp.getX() + rw * .16, aicp.getY() - rh * .015 );
		p2 = new Point2D.Double(aicp.getX() + rw * .36, aicp.getY() - rh * .015 );
		p3 = new Point2D.Double(aicp.getX() + rw * .36, aicp.getY() + rh * .01 );
		p4 = new Point2D.Double(aicp.getX() + rw * .19, aicp.getY() + rh * .01 );
		p5 = new Point2D.Double(aicp.getX() + rw * .19, aicp.getY() + rh * .030 );
		p6 = new Point2D.Double(aicp.getX() + rw * .16, aicp.getY() + rh * .030 );
		p7 = new Point2D.Double(aicp.getX() + rw * .16, aicp.getY() - rh * .015 );
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
		g2.setColor(clrBlack);
		g2.fill(pol);
		g2.setColor(clrWhite);
		g2.draw(pol);
        
        // bank 0 triangle heading indicator
        g2.setStroke( lineStroke );            
		a = new Angle(0);
        p1 = getPoint( aicp, rh * 0.45, 270);
        p2 = getPoint( p1, rh * .05, 60);
        p3 = getPoint (p1, rh * .05, 120);
        pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        g2.setColor( clrWhite );
        g2.fill( pol);
		
		// draw flight director bars
		/*
		if(flightDirOn) {
			g2.setColor(clrPink);
			g2.setStroke(new BasicStroke((int)Math.round(r*0.01)));
			pf = 270;
			p1 = new Point2D.Double(aicp.getX()-r*.25, aicp.getY()+r*fdpitch/pf);
			p2 = new Point2D.Double(aicp.getX()+r*.25, aicp.getY()+r*fdpitch/pf);
			g2.draw(new Line2D.Double(p1,p2));
			p1 = new Point2D.Double(aicp.getX()+r*-fdbank/pf, aicp.getY()-r*.25);
			p2 = new Point2D.Double(aicp.getX()+r*-fdbank/pf, aicp.getY()+r*.25);
			g2.draw(new Line2D.Double(p1,p2));
		}
		*/
		// set old clip
		g2.setClip(oldClip);
		
	}
	
    protected void drawSky( final double rw, final double rh, final Point2D cp, final Point2D aicp, Graphics2D g2 ) 
    {
		Point2D p1, p2, p3, p4;
		Polygon pol;
    	Shape oldClip = g2.getClip();

		// -----------------------------------------------
		// draw attitude indicator scale (sky, scale, etc)
		// -----------------------------------------------
		// sky
		double pf = 45;
		double pv = pitch / pf;
		p1 = new Point2D.Double( cp.getX() - rw, cp.getY() - rh * .845 );
		Rectangle2D.Double rr = new Rectangle2D.Double( p1.getX(), p1.getY(), rw * 2, rh * 1.77 );
		g2.setClip(rr);
		
		Angle a = new Angle( -1 * bank );
		p1 = getPoint( aicp, rh * pv, a.getDegreesPlus(90) );
		p1 = getPoint( p1, rw * 1.8, a.getDegrees() );
		p2 = getPoint( p1, rw * 3.6, a.getDegreesPlus(180));
		p3 = getPoint( p2, rh * 180 / pf, a.getDegreesPlus(90));
		p4 = getPoint( p1, rh * 180 / pf, a.getDegreesPlus(90));
		pol = new Polygon();
	    pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
	    pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
	    pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
	    pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
		
		g2.setColor(skyColorAI);
		g2.fill(pol);
	
		// horizont line
		g2.setColor( new Color( 181, 183, 189 ));
		g2.setStroke(new BasicStroke((int)Math.round(rh*0.0025)));
	    g2.draw( new Line2D.Double(p1, p2));

    	// outer air temperature box
        double delta = rw * 2 / 12; 

    	p1 = new Point2D.Double(cp.getX() - rw, cp.getY() + rh * .86 );
    	Rectangle2D rect = new Rectangle2D.Double( p1.getX(), p1.getY(), delta * 1.5, rh * .06 );
    	g2.setColor( Color.BLACK );
    	g2.fill( rect );
    	g2.setColor( Color.LIGHT_GRAY );
    	p2 = new Point2D.Double(cp.getX() - rw + delta * 1.5, cp.getY() + rh * .86 );
    	g2.draw( new Line2D.Double( p1, p2 ));
    	p1 = new Point2D.Double(cp.getX() - rw + delta * 1.5, cp.getY() + rh * .92 );
    	g2.draw( new Line2D.Double( p1, p2 ));

    	// xpdr / time box
    	p1 = new Point2D.Double(cp.getX() + rw - delta * 3, cp.getY() + rh * .86 );
    	rect = new Rectangle2D.Double( p1.getX(), p1.getY(), rw * delta * 3, rh * .06 );
    	g2.setColor( Color.BLACK );
    	g2.fill( rect );
    	g2.setColor( Color.LIGHT_GRAY );
    	p2 = new Point2D.Double(cp.getX() + rw , cp.getY() + rh * .86 );
    	g2.draw( new Line2D.Double( p1, p2 ));
    	p1 = new Point2D.Double(cp.getX() + rw - delta * 3, cp.getY() + rh * .92 );
    	p2 = new Point2D.Double(cp.getX() + rw - delta * 3, cp.getY() + rh * .86 );
    	g2.draw( new Line2D.Double( p1, p2 ));
    	p1 = new Point2D.Double(cp.getX() + rw - delta * 1.5, cp.getY() + rh * .86 );
    	p2 = new Point2D.Double(cp.getX() + rw - delta * 1.5, cp.getY() + rh * .92 );
    	g2.draw( new Line2D.Double( p1, p2 ));

	    g2.setClip(oldClip);
    }

	protected void drawASI( final double rw, final double rh, final Point2D aicp, Graphics2D g2 ) {
		Point2D p0, p1, p2, p3, p4, p5, p6, p7;
		Polygon pol;
    	Rectangle2D rect;
    	Shape oldClip = g2.getClip();
        double heightfactor = .9;
        double r = rh;
        double dy, ys;
        
		// ---------------------------
        // draw airspeed indicator box
        // ---------------------------
		g2.setStroke(new BasicStroke((int)Math.round(r*0.005)));
		double width = r * .035;
		dy = heightfactor / 120;
		double fps_kts = 0.592483801;
		ys = aicp.getY() + r * ias * dy;

		g2.setColor( Color.LIGHT_GRAY );
		p1 = new Point2D.Double(aicp.getX() - r * .65 - r * .25, aicp.getY() - r * .4 );
		rect = new Rectangle2D.Double( (int)p1.getX(),(int)p1.getY(), (int)(r*.25), (int)( r* heightfactor ) );
		g2.setClip( rect );

		// red bar (0 kts to vs0 speed)
		double h0 = vs0speed.getDoubleValue() * r * fps_kts * dy;
		p1 = new Point2D.Double(aicp.getX() - r * .675, ys - h0  );
        rect = new Rectangle2D.Double( p1.getX(), p1.getY(), width, h0 );
        g2.setColor( new Color( 165, 55, 55 ));
        g2.fill( rect );

		// grey bar (vs0 speed kts to vs1 speed)
		double h1 = vs1speed.getDoubleValue() * r * fps_kts * dy;
		p1 = new Point2D.Double(aicp.getX() - r * .675, ys - ( h0 + h1 ));
        rect = new Rectangle2D.Double( p1.getX(), p1.getY(), width, h1 );
        g2.setColor( new Color( 111, 111, 111 ));
        g2.fill( rect );

		// green bar (vs0 speed kts to vs1 speed)
        double mach_kts = 666.74;
		double hg = maxmach.getDoubleValue() * mach_kts * r * dy;
		p1 = new Point2D.Double(aicp.getX() - r * .675, ys - ( h0 + h1 + hg ));
        rect = new Rectangle2D.Double( p1.getX(), p1.getY(), width, hg );
        g2.setColor( new Color( 88, 148, 88 ));
        g2.fill( rect );
		g2.setClip( oldClip );

		// red bar (max speed)
		double hm = asbarberpole.getDoubleValue() * r * dy;
		p1 = new Point2D.Double(aicp.getX() - r * .695, ys - ( h0 + h1 + hg + hm ));
        rect = new Rectangle2D.Double( p1.getX(), p1.getY(), width, hm );
        g2.setColor( new Color( 165, 55, 55 ));
        g2.fill( rect );
		g2.setClip( oldClip );

		// -----------------------------
        // draw airspeed indicator scale
        // ------------------------------

		g2.setStroke( lineStroke );
		g2.setColor( clrWhite );
		p1 = new Point2D.Double(aicp.getX() - r * .65 - r * .25, aicp.getY() - r * .4 );
		rect = new Rectangle2D.Double( (int)p1.getX(),(int)p1.getY(), (int)(r*.25), (int)( r* heightfactor ) );
		g2.draw( rect );
		g2.setClip( rect );

		p1 = new Point2D.Double(aicp.getX() - r * .65, aicp.getY() - r * .35 );
		rect = new Rectangle2D.Double((int)p1.getX(),(int)p1.getY(), (int)(r*.25), (int)(r*.9));

		for( int i = (int)Math.floor(ias / 10) * 10 - 70; i <= Math.floor(ias / 10) * 10 + 70; i += 10 ) {
			p1 = new Point2D.Double(aicp.getX() - r * .65 , ys - r * dy * i );
			p2 = new Point2D.Double(aicp.getX() - r * ( i % 20 == 0 ? .69 : .675 ) , ys - r * dy * i );
			if( i < 0 ) continue;
			g2.draw( new Line2D.Double(p1, p2)); 
			if( i % 20 == 0 ) {
				p3 =  new Point2D.Double(aicp.getX() - r * .7 , aicp.getY() + r * .045 + r * ias * dy - r * dy * i );
			    drawTextRight( r, p3, g2, .6, ""+i );
			}
		}

	    // --------------------------
	    // Airspeed value display box
	    // --------------------------
		g2.setStroke( lineStroke );

		// draw airspeed box
		p1 = new Point2D.Double(aicp.getX() - r * .65, aicp.getY() );
		p2 = getPoint( p1, r*.03,  135 );
		p3 = getPoint( p2, r*.03 , 90  );
		p4 = getPoint( p3, r*.21 , 180 );
		p5 = getPoint( p4, r*.095, 270 );
		p6 = getPoint( p5, r*.21 , 0   );
		p7 = getPoint( p6, r*.03 , 90  );
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
		g2.setColor(clrBlack);
		g2.fill(pol);
		g2.setColor(clrWhite);
		g2.draw(pol);

		// draw selected airspeed marker (left)
		dy = heightfactor / 120;
		double das = sas - ias;
		das = das < -60 ? -60 : das;
		das = das >  60 ?  60 : das;
		ys = aicp.getY() - r * das * dy;
		p1 = new Point2D.Double(aicp.getX() - r * .65 , ys );
		p2 = getPoint( p1, r*.035, 45 );
		p3 = getPoint( p2, r*.05 , 270);
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        g2.setColor( clrWhite );
		g2.fill(pol);
		
		// draw indicated airspeed (left)
		g2.setColor( clrWhite );
		p1 = new Point2D.Double(aicp.getX() - r * .7, aicp.getY() + r * .05 );
	    drawTextRight( r, p1, g2, .75, ""+Math.round(ias));

		// set old clip
		g2.setClip(oldClip);

	    // draw tas (bottom left)
	    rect = new Rectangle2D.Double( aicp.getX() - r * (0.65 + .25), aicp.getY() + rh * .5, r * .25, rh * .052 );
	    g2.setColor( clrBlack );
	    g2.fill( rect );
	    g2.setColor( clrWhite );
	    g2.draw( rect );
		p1 = new Point2D.Double(aicp.getX() - r * .885, aicp.getY() + rh * .5575 );
	    drawTextLeft( r, p1, g2, .45, "TAS" );
		p1 = new Point2D.Double(aicp.getX() - r * .665, aicp.getY() + rh * .5575 );
	    drawTextRight( r, p1, g2, .45, ""+Math.round( tas ));

	}
	
	protected void drawHSI( final double rw, final double rh, final Point2D hsicp, Graphics2D g2 ) {
		Point2D p0, p1, p2, p3, p4, p5, p6, p7;
		Polygon pol;
    	Rectangle2D rect;
    	Ellipse2D ell;
    	Shape oldClip = g2.getClip();
        double r = rh;
        Font oldFont = g2.getFont();

		// ------------------------------
		// horizontal situation indicator
		// ------------------------------
		
		// draw selected heading marker (bottom)
		Angle a = new Angle( 90 + hi - shdg );
        Point2D pc = new Point2D.Double( hsicp.getX(), hsicp.getY() );

		p0 = new Point2D.Double( pc.getX(), pc.getY() );
        p1 = getPoint( p0, r*.35, a.getDegrees());
		p2 = getPoint( p1, r*.04, a.getDegreesMinus(30));
		p3 = getPoint( p2, r*.02, a.getDegreesMinus(90 ));
		p4 = getPoint( p3, r*.04, a.getDegreesMinus(180));
		p5 = getPoint( p1, r*.04, a.getDegreesPlus(30));
		p6 = getPoint( p5, r*.02, a.getDegreesPlus(90));
		p7 = getPoint( p6, r*.04, a.getDegreesPlus(180));
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
		g2.setColor(new Color( 164,189,200 ));
		g2.fill(pol);

		// draw magnetic heading box + text
        p1 = new Point2D.Double( hsicp.getX() - r * .075, hsicp.getY() - r * .45 );
        rect = new Rectangle2D.Double( p1.getX(), p1.getY(), r * 0.15, r * .07 );
        g2.setColor( Color.BLACK );
        g2.fill( rect );
        g2.setColor( Color.LIGHT_GRAY );
        g2.draw( rect );
        p1 = new Point2D.Double( hsicp.getX(), hsicp.getY() - r * .37 );
        drawText( r, p1, g2, .6, ""+ Math.round( hi ) +"");

		// -----------------------------------------
    	// draw horizontal situation indicator scale
		// -----------------------------------------
        double majorFontSize = Math.max( 6.0, rh * .05);  // minimum 6.0
        Font font = oldFont.deriveFont( Font.PLAIN, (float)majorFontSize );
        g2.setFont( font);
        Angle currentAngle = new Angle( -1 * bank );
        TextLayout stl;
        Shape ts;
    	AffineTransform tat = new AffineTransform();

    	g2.setClip(oldClip);
        pc = new Point2D.Double( hsicp.getX(), hsicp.getY() );
        g2.setColor( new Color( 65, 69, 55 ));
		ell = new Ellipse2D.Double((int)pc.getX()-r*.35,(int)pc.getY()-r*.35, (int)(r*.7), (int)(r*.7));
		g2.fill(ell);
		
        double r1 = r * .32;
        double majorTickSize = proz(r,3);
        double minorTickSize = proz(r,2);
        double majorTickR =  r1 + 2.5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        double minValue = 0;
        double maxValue = 359;
        double minorTickStep = 5;
        double majorTickStep = 30;
        double angleStart = 90;
        double angleExtend = 360;

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( clrWhite );
        	g2.setStroke( new BasicStroke(1.f) );            
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / minorTickStep );
                p0 = getPoint( pc, minorTickR, currentAngle.getDegrees() );
                p1 = getPoint( pc, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){

            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            majorFontSize = Math.max( 6.0, r * .05 * f);  // minimum 6.0
            g2.setColor( clrWhite );
            g2.setStroke( new BasicStroke( (int)(r * 0.0075 )) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                currentAngle = new Angle( 90 - angleStart + angleExtend - stepAngle * (v - hi - minValue) / majorTickStep );

                p0 = getPoint( pc, majorTickR, currentAngle.getDegrees() );
                p1 = getPoint( pc, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s = valueToString( v / 10 );
            	switch((int)(v / 10)) {
            	case  0: s = "N"; break;
            	case  9: s = "E"; break;
            	case 18: s = "S"; break;
            	case 27: s = "W"; break;
            	default: break;
            	}
            	
            	// rotate markers
                stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(r * 0.05)), new FontRenderContext(null, true, false));
            	tat = new AffineTransform();
            	currentAngle = new Angle( 360 - currentAngle.getDegrees() - 135);
            	rect = stl.getBounds();
            	tat.translate(pc.getX() - rect.getWidth()/2, pc.getY() - stl.getAscent()/2 ); 
            	tat.rotate( currentAngle.getRadians(), rect.getCenterX(), rect.getCenterX() );
            	tat.translate( r *.17, r *.17 );
            	tat.rotate( Math.toRadians( 135 ), rect.getCenterX(), rect.getCenterX() );
            	ts = stl.getOutline(tat);
                g2.fill(ts);
                
                // low performance version (markers not rotated) 
                /*
                majorFontSize = r * Math.max( 6.0, r * (v % 30 == 0 ? .07 : .05) * f);  // minimum 6.0
                Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
                Font font = oldFont.deriveFont( Font.PLAIN , Math.round(r * .5 ));
                g2.setFont( font);        
                
                Dimension size = getStringSize( s, g2 );
                double delta = size.getHeight() / 2;
                double fontR = markerR - (v % 30 == 0 ? 3 : 3.25) * majorTickSize;
                Point2D mc = getPoint( pc, fontR - delta, currentAngle.getDegrees() ); // marker center point
                
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
                */
            }
        }	

		// set old clip
		g2.setClip(oldClip);

		double fo = .5, fi = .18;
    	Angle angleCurrent;

    	Ellipse2D.Double ello = new Ellipse2D.Double( hsicp.getX() - r * fo, hsicp.getY() - r * fo, 2 * r * fo, 2 * r * fo );
    	Ellipse2D.Double elli = new Ellipse2D.Double( hsicp.getX() - r * fi, hsicp.getY() - r * fi, 2 * r * fi, 2 * r * fi );

    	// draw ai needle

    	// draw asi ias marker
    	
    	// draw altimeter marker
    	
    	// needle1 (vor2)
    	if( navHasNav2.getBooleanValue() ) {
	        angleCurrent = new Angle( 270 + hi - vor2Radial.getDoubleValue() );
	
	        g2.setColor( new Color( 93, 130, 128 ));
	        g2.setStroke( new BasicStroke((float)(r*.006)));
	        p1 = getPoint( hsicp, r* fo/2 + r * .025, angleCurrent.getDegrees() );
	        p2 = getPoint( p1, r * fo + r * .05, angleCurrent.plus(180) );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .04, angleCurrent.plus(135) );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .04, angleCurrent.plus(-135));
	        g2.draw( new Line2D.Double(p1, p2));
	    	
	        // static needle (in box)
	        p1 = new Point2D.Double( hsicp.getX() - r * fo * .8, hsicp.getY() + r * fo * .6 );
	        p2 = getPoint( p1, r * fo / 5, 0 );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .03, 45 );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .03, 315 );
	        g2.draw( new Line2D.Double(p1, p2));
    	}
        
    	// needle2 (adf1)
    	if( adfSignal.getDoubleValue() > 0.1 ) {
	        angleCurrent = new Angle( 90 - adfr );
	        //angleCurrent = new Angle( 360 - angleCurrent.invert().getDegrees() );

	        g2.setColor( new Color( 93, 130, 128 ));
	        g2.setStroke( new BasicStroke((float)(r*.006)));
	        p1 = getPoint( hsicp, r* fo/2 + r *.05, angleCurrent.getDegrees() );
	        p2 = getPoint( p1, r * .025, angleCurrent.plus(135));
	        p3 = getPoint( p2, r * fo * 1.05, angleCurrent.plus(180));
	        p4 = getPoint( p1, r * .025, angleCurrent.plus(-135));
	        p5 = getPoint( p4, r * fo * 1.05, angleCurrent.plus(180));
	        g2.draw( new Line2D.Double(p4, p5)); // linke pfeilspitze
	        g2.draw( new Line2D.Double(p2, p3)); // rechte pfeispitze
	        g2.draw( new Line2D.Double(p3, p5)); // linke linie
	
	        p2 = getPoint( hsicp, r* fo/2 + r *.05, angleCurrent.plus(180) );
	        p3 = getPoint( p2, r*.05, angleCurrent.getDegrees());
	        g2.draw( new Line2D.Double(p2,p3)); // mittlere linie hinten
	        
	        p2 = getPoint( p1, r * .05, angleCurrent.plus(135) ); // pfeilspitze
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .05, angleCurrent.plus(-135));
	        g2.draw( new Line2D.Double(p1, p2));
	
	        // static needle2
	        p1 = new Point2D.Double( hsicp.getX() + r * fo * .8, hsicp.getY() + r * fo * .6 );
	        p2 = getPoint( p1, r * .015, 225 );
	        p3 = getPoint( p2, r * fo / 5, 180 );
	        g2.draw( new Line2D.Double(p2, p3));
	        p2 = getPoint( p1, r * .03, 225 );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .015, 145 );
	        p3 = getPoint( p2, r * fo / 5.1, 180 );
	        g2.draw( new Line2D.Double(p2, p3));
	        p2 = getPoint( p1, r * .03, 145 );
	        g2.draw( new Line2D.Double(p1, p2));
    	}
    	
        // cdi needle
        // outer part (top and bottom) / todo: setClip
    	Area area = new Area( ello );
    	g2.setClip(area);
    	
        Color clrNeedle = null;
        double deflection = cdi;
        if( gpsDrivesNav1.getBooleanValue() && hsiBearingValid.getBooleanValue() ) {
            angleCurrent = new Angle( 90 + hi - hsiBearing.getDoubleValue() );
            deflection = hsiCDI.getDoubleValue();
	        clrNeedle = clrPink;
        }
//        else if( gpsDrivesNav1.getDoubleValue() == 1  && (Double)hsiBearingValid.getValue() == 1.0 ) {
        	
//        }
        else {
            angleCurrent = new Angle( 90 + hi - obs );
            clrNeedle = clrGreen;
        }
        if( clrNeedle != null ) {
	        g2.setColor( clrNeedle );
	        g2.setStroke( new BasicStroke((float)(r*.01)));
	        p1 = getPoint( hsicp, r* fo/2, angleCurrent.getDegrees() );
	        p2 = getPoint( p1, r * fo, angleCurrent.plus(180) );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .05, angleCurrent.plus(135) );
	        g2.draw( new Line2D.Double(p1, p2));
	        p2 = getPoint( p1, r * .05, angleCurrent.plus(-135));
	        g2.draw( new Line2D.Double(p1, p2));
        }
        
        g2.setColor( new Color( 55, 69, 55 ));
        g2.fill(elli);
        g2.setColor( clrWhite );
        g2.setStroke( new BasicStroke((float)(r*.005)));
        g2.draw(elli);

        // inner scale
        g2.setColor( clrWhite );
        g2.setStroke( new BasicStroke((float)(r*.005)));
        double tickspace = fi * .8 / 2;
        for( double i = -2 * tickspace; i <= tickspace * 2; i += tickspace ) {
        	if( i == 0 ) continue;
            p2 = getPoint( hsicp, r * i, angleCurrent.plus(90) );
            ell = new Ellipse2D.Double( p2.getX(), p2.getY(), r * 0.01, r * 0.01 );
            g2.draw( ell );
        }

        if( clrNeedle != null ) {
	        g2.setColor( clrNeedle );
	        g2.setStroke( new BasicStroke((float)(r*.01)));
	        p1 = getPoint( hsicp, r * fi *.8 * deflection / 128, angleCurrent.plus(270) );
	        p2 = getPoint( p1, r * fi * 1.5 /2, angleCurrent.getDegrees());
	        p1 = getPoint( p2, r * fi* 1.5 , angleCurrent.plus(180));
	        g2.draw( new Line2D.Double(p1, p2));
        }
        
		// set old clip
		g2.setClip(oldClip);

	}
	
	protected void drawALT( final double rw, final double rh, final Point2D aicp, Graphics2D g2 ) {
		Point2D p0, p1, p2, p3, p4, p5, p6, p7;
        Ellipse2D ell;
		Polygon pol;
    	Rectangle2D rect;
    	Shape oldClip = g2.getClip();
		double r = rh;
        double heightfactor = .9;
        double ys, dy;
        // --------------------
		// altimeter rect right
		// --------------------
		g2.setClip(oldClip);
		
		p1 = new Point2D.Double(aicp.getX() + r * .65, aicp.getY() - r * .4 );
		rect = new Rectangle2D.Double( (int)p1.getX(),(int)p1.getY(), (int)(r*.25), (int)( r* heightfactor ) );
		g2.draw( rect );
		g2.setClip( rect );

    	// draw altimeter scale
        p1 = new Point2D.Double(aicp.getX() + r * .65, aicp.getY() - r * .4 );
		rect = new Rectangle2D.Double((int)p1.getX(),(int)p1.getY(), (int)(r*.25), (int)(r * heightfactor));
		g2.setClip(rect);
		g2.setColor( Color.LIGHT_GRAY );

		dy = .9 / 800;
		ys = aicp.getY() + r * ialt * dy;
		for( int i = (int)Math.floor(ialt / 100) * 100 - 400; i <= Math.floor(ialt / 100) * 100 + 400; i += 100 ) {
			p1 = new Point2D.Double(aicp.getX() + r * .65 , ys - r * dy * i );
			p2 = new Point2D.Double(aicp.getX() + r * .69 , ys - r * dy * i );
			if( i < -1000 ) continue;
			g2.draw( new Line2D.Double(p1, p2)); 
			if( i % 200 == 0 ) {
				p3 =  new Point2D.Double(aicp.getX() + r * .77 , aicp.getY() + r * .045 + r * ialt * dy - r * dy * i );
			    drawText( r, p3, g2, .5, ""+i );
			}
		}

		// draw selected altitude (top right)
		p1 = new Point2D.Double(aicp.getX(), aicp.getY() - r * .25 );
	    drawTextRight( r, p1, g2, .6, ""+Math.round(salt));

	    // draw selected altitude marker (right)
		dy = heightfactor  / 800;
		double das = salt- ialt;
		das = das < -400 ? -400 : das;
		das = das >  400 ?  400 : das;
		ys = aicp.getY() - r * .3 - r * das * dy;

		p1 = new Point2D.Double(aicp.getX() + r * .66 , ys );
		p2 = getPoint( p1, r*.03 , 45 );
		p3 = getPoint( p2, r*.02 , 90 );
		p4 = getPoint( p3, r*.05 , 180);
		p5 = getPoint( p4, r*.082, 270);
		p6 = getPoint( p5, r*.05 , 0  );
		p7 = getPoint( p6, r*.02 , 90 );
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
		g2.setColor( new Color (159, 186, 195 ));
		g2.fill(pol);

		g2.setClip(oldClip);

	    // ---------------------
	    // RIGHT RECT - ALTITUDE
	    // ---------------------

		// draw altimeter box
		p1 = new Point2D.Double(aicp.getX() + r * .65, aicp.getY() );
		p2 = getPoint( p1, r*.03, 45);
		p3 = getPoint( p2, r*.03 , 90  );
		p4 = getPoint( p3, r*.21 , 0);
		p5 = getPoint( p4, r*.095, 270);
		p6 = getPoint( p5, r*.21 , 180  );
		p7 = getPoint( p6, r*.03 , 90 );
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
		g2.setColor(clrBlack);
		g2.fill(pol);
		g2.setColor(clrWhite);
		g2.draw(pol);

		// draw indicated altitude (right)
		g2.setColor( clrWhite );
		p1 = new Point2D.Double(aicp.getX() + r * .88, aicp.getY() + r *.05 );
		int ia = (int)Math.round(ialt/20)*20;
	    drawTextRight( r, p1, g2, .65, ""+ia);

    	// draw qnh (bottom right)
	    rect = new Rectangle2D.Double( aicp.getX() + r * 0.65, aicp.getY() + rh * .5, r * .25, rh * .052 );
	    g2.setColor( clrBlack );
	    g2.fill( rect );
	    g2.setColor( clrWhite );
	    g2.draw( rect );
		p1 = new Point2D.Double(aicp.getX() + r * .775 , aicp.getY() + rh * .5575 );
	    drawText( r, p1, g2, .45, Math.round(qnh) +" mb");

	    // draw selected altitude (top right)
	    rect = new Rectangle2D.Double( aicp.getX() + r * 0.65, aicp.getY() - rh * .45, r * .25, rh * .052 );
	    g2.setColor( clrBlack );
	    g2.fill( rect );
	    g2.setColor( clrWhite );
	    g2.draw( rect );
		p1 = new Point2D.Double(aicp.getX() + r * .775 , aicp.getY() - r * .39 );
	    drawText( r, p1, g2, .45, Math.round(salt) +"" );

        // ----------
        // variometer
        // ----------
    	g2.setStroke( lineStroke );

        double vsi = vs > 2000 ? 2000 : vs; vsi = vsi < -2000 ? -2000 : vsi;
        double vsf = ( .225 / 1.5 ) * (vsi / 1000);
		p1 = new Point2D.Double(aicp.getX() + r * .65, aicp.getY() );
        if( vsi > 0 ) {
        	rect = new Rectangle2D.Double( p1.getX(), p1.getY() - vsf * r, r * .02, vsf * r );
        } else {
        	rect = new Rectangle2D.Double( p1.getX(), p1.getY() * r, r * .02, -vsf * r );
        }
        g2.setColor( new Color( 142, 135, 151 ));
        g2.fill( rect );

        // border
        g2.setColor( Color.LIGHT_GRAY );
        p1 = getPoint( aicp, r * 0.9, 0 );
        p2 = getPoint( p1, r * .1, 25 );
        p3 = getPoint( p1, r * .1, 335 );
        drawLine( p1, p2, g2 );
        drawLine( p1, p3, g2 );
        p4 = getPoint( p2, r * .3, 90 );
        p5 = getPoint( p3, r * .3, 270 );
        drawLine( p2, p4, g2 );
        drawLine( p3, p5, g2 );
        p2 = getPoint( p4, r * .09, 180 );
        p3 = getPoint( p5, r * .09, 180 );
        drawLine( p2, p4, g2 );
        drawLine( p3, p5, g2 );
        // scale
        double h = .225 / 3 * r;
        int len;
        for( int i = 0; i <= 4; i++ ) {
        	if( i == 0 ) continue;
        	if( i % 2 == 0 ) {
        		len = (int)Math.round( r * .04 );
        		p1 = new Point2D.Double(aicp.getX() + r * .95, aicp.getY() - i * h + r * .0375 );
        		drawTextLeft( r, p1, g2, .5, ""+ (i / 2));
				p1 = new Point2D.Double(aicp.getX() + r * .95, aicp.getY() + i * h + r * .0375 );
				drawTextLeft( r, p1, g2, .5, ""+ (i / 2));
			}
        	else  { 
        		len = (int)Math.round( r * .02 );
        	}
    		p1 = new Point2D.Double(aicp.getX() + r * .9, aicp.getY() - i * h );
    		p2 = new Point2D.Double(aicp.getX() + r * .9 + len, aicp.getY() - i * h );
            drawLine( p1, p2, g2 );
    		p1 = new Point2D.Double(aicp.getX() + r * .9, aicp.getY() + i * h );
    		p2 = new Point2D.Double(aicp.getX() + r * .9 + len, aicp.getY() + i * h );
            drawLine( p1, p2, g2 );
        }
        // box
		p1 = new Point2D.Double(aicp.getX() + r * .9, aicp.getY() - vsf * r );
        p2 = getPoint( p1, r * .07, 25  );
        p3 = getPoint( p2, r * .12, 0   );
        p4 = getPoint( p3, r * .059, 270 );
        p5 = getPoint( p4, r * .12, 180 );
        pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
        g2.setColor( Color.black );
        g2.fill( pol );
        g2.setColor( Color.LIGHT_GRAY );
        g2.draw( pol );
        // value
		p1 = new Point2D.Double(aicp.getX() + r * 1.08, aicp.getY() - vsf * r + r * .04);
		String s = Math.round( vs * 2 / 100 ) * 10 / 2 +"0";
		if( s.equals("00")) s = "";
        drawTextRight( r, p1, g2, .5, s );


        // ----------------
        // glidescope scale
        // ----------------
    	g2.setStroke( lineStroke );

        if( gsf == 1.0 ) {
        
	        // border
	        g2.setColor( Color.LIGHT_GRAY );
	        h = r * .225;
			p1 = new Point2D.Double(aicp.getX() + r * .645, aicp.getY() - h );
	        p2 = getPoint( p1, r * .075, 180 );
	        drawLine( p1, p2, g2 );
	        p3 = getPoint( p2, 2 * h, 270 );
	        drawLine( p2, p3, g2 );
	        p4 = getPoint( p3, r * .075, 0 );
	        drawLine( p3, p4, g2 );

	        // scale
	        for( int i = -2; i <= 2; i++ ) {
	        	if( i == 0 ) continue;
	    		p1 = new Point2D.Double(aicp.getX() + r * .595, aicp.getY() - r * .014 - i * h / 2.5);
	        	ell = new Ellipse2D.Double( p1.getX(), p1.getY(), r * .027, r * .028 );
	        	g2.draw( ell );
	        }
			p1 = new Point2D.Double(aicp.getX() + r * .645, aicp.getY() );
	        p2 = getPoint( p1, r * .075, 180 );
	        drawLine( p1, p2, g2 );

	        // needle
	        double gsi = gsa > 1 ? 1 : gsa; gsi = gsi < -1 ? -1 : gsi; 
	        double gsy = r * ( .225 / 2.5 ) * gsi * 2;
			p1 = new Point2D.Double(aicp.getX() + r * .595 + r * .042, aicp.getY() + gsy );
	        p2 = getPoint( p1, r * .04, 135 );
	        p3 = getPoint( p2, r * .04, 135 + 90 );
	        p4 = getPoint( p3, r * .04, 135 + 180 );
	        
	        pol = new Polygon();
	        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
	        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
	        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
	        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
	        g2.setColor( new Color( 113, 163, 104 ));
	        g2.fill( pol );
        }

	}
	
    protected void drawText( final double rw, final double rh, final Point2D cp, final Point2D aicp, final Point2D hsicp, Graphics2D g2 ) {
		Point2D p1, p2;
		Ellipse2D ell;
        double r = rh;

        // ---------------
		// nav frequencies
        // ---------------
        Rectangle2D rr;

        // blue box
        if( editNav == 1 ) {
            // upper (1)
			p1 = new Point2D.Double(cp.getX() - rw * .905, cp.getY() - rh * .985 );
			rr = new Rectangle2D.Double( p1.getX(), p1.getY(), r * .18, r * .06);
        }
        else {
	        // lower (2)
			p1 = new Point2D.Double(cp.getX() - rw * .905, cp.getY() - r * .92 );
			rr = new Rectangle2D.Double( p1.getX(), p1.getY(), r * .2, r * .06);
        }
		g2.setColor( new Color( 44, 52, 234 ));
		g2.draw( rr );
		
        // freqs
        g2.setColor( clrWhite );
        double freqspace = 1.175 - .92;
		p1 = new Point2D.Double(cp.getX() - rw * .85, cp.getY() - rh * .92 );
        drawText( r, p1, g2, .5, ""+ freqNF.format(varNav1Stby.getDoubleValue()) );
		p1 = new Point2D.Double(cp.getX() - rw * .85, cp.getY() - rh * .85 );
        drawText( r, p1, g2, .5, ""+ freqNF.format(varNav2Stby.getDoubleValue()) );
        p1 = new Point2D.Double(cp.getX() - rw * .73, cp.getY() - rh * .92 );
        g2.setColor( navHasNav1.getBooleanValue() ? clrGreen : clrWhite );
        drawText( r, p1, g2, .5, ""+ freqNF.format(varNav1Freq.getDoubleValue()) );
		p1 = new Point2D.Double(cp.getX() - rw * .73, cp.getY() - rh * .85 );
        g2.setColor( navHasNav2.getBooleanValue() ? clrGreen : clrWhite );
        drawText( r, p1, g2, .5, ""+ freqNF.format(varNav2Freq.getDoubleValue()) );

        // ---------------
        // com frequencies
        // ---------------
        // blue box
        if( editCom == 1 ) {
            // upper (1)
			p1 = new Point2D.Double(cp.getX() + rw * .755, cp.getY() - rh * .985 );
			rr = new Rectangle2D.Double( p1.getX(), p1.getY(), r * .18, r * .06);
        } else {
        	// lower (2)
			p1 = new Point2D.Double(cp.getX() + rw * 1.065, cp.getY() - rh * .92 );
			rr = new Rectangle2D.Double( p1.getX(), p1.getY(), r * .18, r * .06);
        }
		g2.setColor( new Color( 44, 52, 234 ));
		g2.draw( rr );

        // freqs
        g2.setColor( clrWhite );
        double freq = varCom1Stby.getDoubleValue();
		p1 = new Point2D.Double(cp.getX() + rw * .85, cp.getY() - rh * .92 );
        drawTextRight( r, p1, g2, .5, "1"+(freqNF.format((float)Integer.parseInt(Integer.toHexString((int)freq)) / 100)) );
        freq = varCom2Stby.getDoubleValue();
		p1 = new Point2D.Double(cp.getX() + rw * .85, cp.getY() - rh * .85 );
        drawTextRight( r, p1, g2, .5, "1"+(freqNF.format((float)Integer.parseInt(Integer.toHexString((int)freq)) / 100)) );

        freq = varCom1Freq.getDoubleValue();
		p1 = new Point2D.Double(cp.getX() + rw * .73 - freqspace, cp.getY() - rh * .92 );
        g2.setColor( com1Xmit.getBooleanValue() || comBoth.getBooleanValue() ? clrGreen : clrWhite );
        drawTextRight( r, p1, g2, .5, "1"+(freqNF.format((float)Integer.parseInt(Integer.toHexString((int)freq)) / 100)) );
        freq = varCom2Freq.getDoubleValue();
		p1 = new Point2D.Double(cp.getX() + rw * .73 - freqspace, cp.getY() - rh * .85 );
        g2.setColor( com2Xmit.getBooleanValue() || comBoth.getBooleanValue() ? clrGreen : clrWhite );
        drawTextRight( r, p1, g2, .5, "1"+(freqNF.format((float)Integer.parseInt(Integer.toHexString((int)freq)) / 100)) );

        // ---------------------------------------------
        // upper values (gps waypoint, distance, tracks)
        // ---------------------------------------------
        // 		double gpsgmt, gpswpd, gpswdt; String gpswni;
        g2.setColor( clrPink );
    	String[] txt = { gpswni, distNF.format(gpswd)+" NM", Math.round(gpswdt)+"", Math.round(gpsgmt)+""}; // {"WPT", "DIS", "DTK", "TRK"};
        double delta = rw * 1.3 / 4; // ~ 0.37
    	p1 = new Point2D.Double(cp.getX() - rw * .5, cp.getY() - rh * .925 );
    	drawTextLeft( r, p1, g2, .45, txt[0] );
    	p1 = new Point2D.Double(cp.getX() - rw * .5 + delta, cp.getY() - rh * .925 );
    	drawTextLeft( r, p1, g2, .45, txt[1] );
    	p1 = new Point2D.Double(cp.getX() - rw * .5 + delta * 2, cp.getY() - rh * .925 );
    	drawTextLeft( r, p1, g2, .45, txt[2] );
    	p1 = new Point2D.Double(cp.getX() - rw * .5 + delta * 3, cp.getY() - rh * .925 );
    	drawTextLeft( r, p1, g2, .45, txt[3] );

    	// -------
    	// heading
    	// -------
    	g2.setStroke( lineStroke );

    	double tw = r * .23;
    	if( System.currentTimeMillis() - headingChanged < 3000 ) {
	    	p1 = new Point2D.Double( cp.getX() - r * .3 - tw, cp.getY() + r * .055 );
	    	rr = new Rectangle2D.Double( p1.getX(), p1.getY(), tw, r * .07 );
	    	g2.setColor( clrBlack );
	    	g2.fill( rr );
	    	g2.setColor( clrWhite );
	    	g2.draw( rr );
	    	p2 = new Point2D.Double( p1.getX() + r * .01, p1.getY() + r * .06 );
	    	drawTextLeft( r, p2, g2, .4, "HDG" );
	    	g2.setColor( clrBlue );
	    	p2 = new Point2D.Double( p1.getX() - r * .01 + tw , p1.getY() + r * .07 );
	    	drawTextRight( r, p2, g2, .5, Math.round( headingLockDir.getDoubleValue() ) +"");
    	}

    	// -------
    	// course
    	// -------
    	if( System.currentTimeMillis() - obsChanged < 3000 ) {
	    	p1 = new Point2D.Double( cp.getX() + r * .3, cp.getY() + r * .055 );
	    	rr = new Rectangle2D.Double( p1.getX(), p1.getY(), tw, r * .07 );
	    	g2.setColor( clrBlack );
	    	g2.fill( rr );
	    	g2.setColor( clrWhite );
	    	g2.draw( rr );
	    	p2 = new Point2D.Double( p1.getX() + r * .01  , p1.getY() + r * .06 );
	    	drawTextLeft( r, p2, g2, .4, "CRS" );
	    	g2.setColor( clrGreen );
	    	p2 = new Point2D.Double( p1.getX() - r * .01 + tw , p1.getY() + r * .07 );
	    	drawTextRight( r, p2, g2, .5, Math.round( obs ) +"");
    	}

    	// ------------
    	// lower values
    	// ------------
        delta = r * 2.8 / 12; 

    	// outer air temperature
    	g2.setColor( Color.LIGHT_GRAY );
    	p1 = new Point2D.Double(cp.getX() - rw * .95, cp.getY() + r * .92 );
    	drawTextLeft( r, p1, g2, .45, "OAT "+ Math.round( varOAT.getDoubleValue() ) +"C" );
    	
    	// xpdr code
    	p1 = new Point2D.Double(cp.getX() + rw * .95 - 3 * delta, cp.getY() + r * .92 );
    	drawTextLeft( r, p1, g2, .45, "XPDR");
    	double xpdr = varXpdrCode.getDoubleValue();
    	p1 = new Point2D.Double(cp.getX() + rw * .95 - 1.5 * delta, cp.getY() + r * .92 );
    	g2.setColor( new Color( 100, 150, 100 ));
    	drawTextRight( r, p1, g2, .45, lz(Integer.parseInt(Integer.toHexString((int)xpdr)), 4 )+"ALT");
    	
    	// local time
    	g2.setColor( Color.LIGHT_GRAY );
    	p1 = new Point2D.Double(cp.getX() + rw * .98 - 1.5 * delta, cp.getY() + r * .92 );
    	double lt = varLocalTime.getDoubleValue();
		int hours   = (int)(lt / 3600);
		int minutes = (int)(lt / 60 % 60);
		int seconds = (int)(lt % 60);
    	drawTextLeft( r, p1, g2, .45, "LCL  "+lz(hours,2)+":"+lz(minutes,2)+":"+lz(seconds,2));

    	// ---------------------------
        // marker
        // ---------------------------
    	if( markerI.getBooleanValue() || markerO.getBooleanValue() || markerM.getBooleanValue() ) {
	    	double w = r * .07;
	    	p1 = new Point2D.Double( aicp.getX() +  r * .45, aicp.getY() - r * .4 );
	    	rr = new Rectangle2D.Double( p1.getX(), p1.getY(), w, w );
	        g2.setColor( new Color( 197, 199, 123) );
	        g2.fill( rr );
	        g2.setColor( Color.BLACK );
	        g2.draw( rr );
	        String str = "X"; 
	        str = markerI.getBooleanValue() ? "I" : str; 
		    str = markerM.getBooleanValue() ? "M" : str; 
		    str = markerO.getBooleanValue() ? "O" : str; 
		    p1 = new Point2D.Double( p1.getX() + w/2, p1.getY() + w );
	        g2.setColor( Color.BLACK );
		    drawText( r, p1, g2, .5, str );
    	}
    	
    	// ---------------------------
        // vor1 bearing text box
        // ---------------------------
    	Area ar;
    	if( navHasNav1.getBooleanValue() ) {
	        p1 = new Point2D.Double( hsicp.getX() - r * .65, hsicp.getY() - r * 0.055 );
	    	rr = new Rectangle2D.Double(p1.getX(), p1.getY(), r * .75, r * .2 );
			ell = new Ellipse2D.Double((int)hsicp.getX()-r*.4,(int)hsicp.getY()-r*.4, (int)(r*.8), (int)(r*.8));
			ar = new Area( rr );
			ar.subtract( new Area( ell ));
	        g2.setColor( Color.BLACK );
	        g2.fill( ar );
	        g2.setColor( Color.LIGHT_GRAY );
	        g2.draw( ar );
	
		    // draw text vor1: dme, ident, type 
	        g2.setColor( Color.LIGHT_GRAY );
		    if( hasDME1.getBooleanValue() ) {
				p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * .02 );
				double d = navDME1.getDoubleValue();
				if ( d >= 100 )
				    drawTextLeft( r, p1, g2, .5, Math.round(d) +" NM");
				else 
				    drawTextLeft( r, p1, g2, .5, dmeNF.format(d) +" NM");
		    }
    	
		    boolean gdn1 = gpsDrivesNav1.getBooleanValue();
		    g2.setColor( gdn1 ? new Color( 168, 108, 167 ) : new Color( 108, 168, 107 ));
			p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * ( .02 + .06 ));
		    drawTextLeft( r, p1, g2, .5, ( gpsDrivesNav1.getBooleanValue() ? hsiID.getStringValue() : "VOR1"));
		    
	        g2.setColor( Color.LIGHT_GRAY );
			p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * ( .02 + .12 ));
		    drawTextLeft( r, p1, g2, .5, nav1ident.getStringValue() );
    	}


    	// ---------------------------
        // vor2
        // ---------------------------
    	if( navHasNav2.getBooleanValue() ) {
	        p1 = new Point2D.Double( hsicp.getX() - r * .65, hsicp.getY() + r * .175 );
	    	rr = new Rectangle2D.Double(p1.getX(), p1.getY(), r * .65, r * .2 );
			ell = new Ellipse2D.Double((int)hsicp.getX()-r*.4,(int)hsicp.getY()-r*.4, (int)(r*.8), (int)(r*.8));
			ar = new Area( rr );
			ar.subtract( new Area( ell ));
	        g2.setColor( Color.BLACK );
	        g2.fill( ar );
	        g2.setColor( Color.LIGHT_GRAY );
	        g2.draw( ar );
	
		    // draw text vor2: dme, ident, type 
		    g2.setColor( Color.LIGHT_GRAY );
		    if( hasDME2.getBooleanValue()) {
				p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * .245 );
				double d = navDME2.getDoubleValue();
				if ( d >= 100 )
				    drawTextLeft( r, p1, g2, .5, Math.round(d) +" NM");
				else 
				    drawTextLeft( r, p1, g2, .5, dmeNF.format(d) +" NM");
		    }
	
		    g2.setColor( clrBlue );
			p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * ( .245 + .06 ));
		    drawTextLeft( r, p1, g2, .5, nav2ident.getStringValue());
		    
	        g2.setColor( clrWhite );
			p1 = new Point2D.Double(hsicp.getX() - r * .63, hsicp.getY() + r * ( .245 + .12 ));
		    drawTextLeft( r, p1, g2, .5, "NAV2" );
    	}
    	
    	// ---------------------------
        // adf
        // ---------------------------
    	if( adfSignal.getDoubleValue() > 0.1 ) {
    		// box
	        p1 = new Point2D.Double( hsicp.getX(), hsicp.getY() + r * .175 );
	    	rr = new Rectangle2D.Double(p1.getX(), p1.getY(), r * .65, r * .2 );
			ell = new Ellipse2D.Double((int)hsicp.getX()-r*.4,(int)hsicp.getY()-r*.4, (int)(r*.8), (int)(r*.8));
			ar = new Area( rr );
			ar.subtract( new Area( ell ));
	        g2.setColor( Color.BLACK );
	        g2.fill( ar );
	        g2.setColor( Color.LIGHT_GRAY );
	        g2.draw( ar );

	        // draw text adf: freq, ident, type 
		    g2.setColor( clrWhite );
			p1 = new Point2D.Double(hsicp.getX() + r * .63, hsicp.getY() + r * .245 );
			freq = adfFreq.getDoubleValue();
		    drawTextRight( r, p1, g2, .5, ""+lz((float)Integer.parseInt(Integer.toHexString((int)freq)) / 10000, 6 ));
	
		    g2.setColor( clrBlue );
			p1 = new Point2D.Double(hsicp.getX() + r * .63, hsicp.getY() + r * ( .245 + .06 ));
		    drawTextRight( r, p1, g2, .5, adfIdent.getStringValue() );
		    
	        g2.setColor( clrWhite );
			p1 = new Point2D.Double(hsicp.getX() + r * .63, hsicp.getY() + r * ( .245 + .12 ));
		    drawTextRight( r, p1, g2, .5, "ADF1" );
    	}
    	
    	// -----------------
    	// bottom button bar
    	// -----------------
        // bottom button bar
    	p1 = new Point2D.Double(cp.getX() - rw, cp.getY() + rh * .92 );
    	p2 = new Point2D.Double(cp.getX() + rw, cp.getY() + rh * .92 );
    	g2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );

        String[] buttons = {"", "INSET", "", "PFD", "", "CDI", "", "XPDR", "IDENT", "TMR/REF", "NRST", "ALERTS", ""};
    	g2.setColor( Color.LIGHT_GRAY );
        delta = rw * 2 / 12; 
        for( int i = 1; i < 12; i++ ) {
        	p1 = new Point2D.Double(cp.getX() - rw + i * delta + delta/2, cp.getY() + r * .99 );
        	drawText( r, p1, g2, .45, buttons[i] );

        	p1 = new Point2D.Double(cp.getX() - rw + i * delta, cp.getY() + rh * .92 );
        	p2 = new Point2D.Double(cp.getX() - rw + i * delta, cp.getY() + rh * .99 );
    		g2.drawLine( (int)p1.getX(), (int)p1.getY(), (int)p2.getX(), (int)p2.getY() );
        }
    }

}
