package org.gaugebook.gauges.aviation.glass;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Transparency;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.VolatileImage;
import java.text.NumberFormat;

import org.gaugebook.AbstractGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;

public class MFD737 extends AbstractGauge {

	private static final long serialVersionUID = 1L;
	
	private FSXVariable headingIndicator, headingLockDir;
	private FSXVariable vorOBS, vorCDI, vorToFrom, vorGSAngle, vorGSFlag;
	private FSXVariable groundSpeed, trueAirspeed;

	private FSXVariable nav1ident, nav1name;
	private FSXVariable nav2ident, nav2name;
	private FSXVariable nav1obs, nav2obs;
	private FSXVariable nav1signal, nav2signal;
	private FSXVariable hasDME1, hasDME2, navDME1, navDME2;
	private FSXVariable windVelocity, windDirection; 
	private static NumberFormat dmeNF  = NumberFormat.getInstance();

	private FSXVariable varAviCircuitOn;

	private static final Color clrBlack = new Color(0,0,0);
	private static final Color clrWhite = new Color(255,255,255);
	private static final Color clrGrey = new Color(99,99,99);
	private static final Color clrGreen = new Color(98,246,106);
	private static final Color clrPink = new Color(222,0,222);
	private static final Color clrRed = new Color(222,0,0);

	private double 	gs,			// ground speed
					tas,		// true airspeed
					hi,			// heading indicated
					hld,		// heading lock dir
					shdg,		// selected heading
					hdg,		// magnetic heading
					ws,			// wind speed	
					wd			// wind direction
					;

	private boolean aviOn = false, aviOld = false, firstCall = true;

	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );

		groundSpeed = owner.registerFSXVariable("GPS GROUND SPEED");
		trueAirspeed = owner.registerFSXVariable("AIRSPEED TRUE");
	    headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");
        headingLockDir = owner.registerFSXVariable("AUTOPILOT HEADING LOCK DIR");
        vorOBS = owner.registerFSXVariable("NAV OBS:1");
        vorCDI = owner.registerFSXVariable("NAV CDI:1");
        vorToFrom = owner.registerFSXVariable("NAV TOFROM:1");
		vorGSAngle = owner.registerFSXVariable("NAV GLIDE SLOPE ERROR:1");
		vorGSFlag = owner.registerFSXVariable("NAV GS FLAG:1");

		nav1ident = owner.registerFSXVariable("NAV IDENT:1");
		nav1name = owner.registerFSXVariable("NAV NAME:1");
		nav2ident = owner.registerFSXVariable("NAV IDENT:2");
		nav2name = owner.registerFSXVariable("NAV NAME:2");
		nav1obs = owner.registerFSXVariable("NAV OBS:1");
		nav2obs = owner.registerFSXVariable("NAV OBS:2");
		
		hasDME1 = owner.registerFSXVariable("NAV HAS DME:1");
		hasDME2 = owner.registerFSXVariable("NAV HAS DME:2");
		navDME1 = owner.registerFSXVariable("NAV DME:1");
		navDME2 = owner.registerFSXVariable("NAV DME:2");

		nav1signal = owner.registerFSXVariable("NAV HAS NAV:1");
		nav2signal = owner.registerFSXVariable("NAV HAS NAV:2");
		
		windVelocity = owner.registerFSXVariable("AMBIENT WIND VELOCITY");
		windDirection = owner.registerFSXVariable("AMBIENT WIND DIRECTION");

		varAviCircuitOn = owner.registerFSXVariable("CIRCUIT AVIONICS ON");

		dmeNF.setMaximumFractionDigits(1);
		dmeNF.setMinimumFractionDigits(1);
		dmeNF.setMaximumIntegerDigits(2);
		dmeNF.setMinimumIntegerDigits(1);
			
		initdone = true;
	}

	public void update() {
		if( groundSpeed == null ) return;
		
		gs = groundSpeed.getDoubleValue();
		tas = trueAirspeed.getDoubleValue();
		hi = ( 180 / StrictMath.PI ) * headingIndicator.getDoubleValue();
		hld = headingLockDir.getDoubleValue();
		hdg = headingIndicator.getDoubleValue();
		shdg = headingLockDir.getDoubleValue();
		
		ws = windVelocity.getDoubleValue();
		wd = windDirection.getDoubleValue();
		
		aviOn = varAviCircuitOn.getDoubleValue() == 1;
		repaint();
	}


	int lastWidth = 0, lastHeight = 0;
	
	public void paint( Graphics g ) {
        // super.paintComponent( g );
        int w = getWidth();
        int h = getHeight();

    	Graphics2D g2 = (Graphics2D)g;
        g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));

        double r =  Math.min( w, h ) / 2 - 5;            // radius
        if( r < 0 ) r = 100;
        Point2D cp = new Point2D.Double( w / 2, h / 2 ); 
        drawBackground( r, cp, g2 );
        if( aviOn ) {
            double r0 = proz(r,55);
            double r1 = proz(r,58);
            double majorTickSize = proz(r,4);
            double minorTickSize = proz(r,2);
            double majorTickR =  r1 + 4; // major tick radius 
            double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
            double needleR = majorTickR + 1.5 * majorTickSize;
            drawScale(r,cp,g2);
            drawText(r,cp,g2);
            drawValue(r,needleR, cp,g2);
            drawNeedle( r, needleR, cp, g2 );
        }
        g2.dispose();
    }	
    
    protected void drawBackground( final double r, final Point2D cp, Graphics2D g2 ) {
		Point2D p0, p1, p2, p3;
		Polygon pol;
		Ellipse2D ell;
		boolean update = firstCall;
		int w = getWidth();
		int h = getHeight();

        GraphicsConfiguration gc = getGraphicsConfiguration();
    	if( update || backVi != null && backVi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
    		//firstCall = false;
    		backVi = gc.createCompatibleVolatileImage( getWidth(), getHeight() );
            
	        do {
	        	gc = this.getGraphicsConfiguration();
	        	if( backVi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
	        		backVi = gc.createCompatibleVolatileImage( getWidth(), getHeight() );
	        	}
	        	
	        	Graphics2D bg2 = backVi.createGraphics();
	            bg2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));

				bg2.setClip( 0, 0, w, h );
		        bg2.setColor( Color.DARK_GRAY );
		        bg2.fillRect( 0, 0, w, h );
				bg2.setColor( clrBlack );
	
	            // some 3d bevel would be nice here
				bg2.setColor(clrBlack);
				p1 = new Point2D.Double(cp.getX() - r * .98, cp.getY() - r * 0.95 );
		    	RoundRectangle2D.Double rr = new RoundRectangle2D.Double(p1.getX(), p1.getY(), r * 1.9, r * 1.8, r * .1, r *.1 );
				bg2.fill(rr);
				
				// return if avionics circuit if off
				if( !aviOn ) return;
				
				// heading rect
				p1 = new Point2D.Double(cp.getX() - r*.1, cp.getY() - r * .93 );
		    	Rectangle2D.Double rect = new Rectangle2D.Double(p1.getX(), p1.getY(), r * .2, r * .1 );
		    	bg2.setColor(clrWhite);
		    	bg2.draw(rect);
	
		        // plane triangle
		        p1 = getPoint( cp, +r * .54, 270);
		        p2 = getPoint( p1, r * .15, 250);
		        p3 = getPoint (p1, r * .15, 290);
		        pol = new Polygon();
		        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
		        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
		        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
		        bg2.setColor( clrWhite );
		        bg2.draw( pol);
	
		        // vertical line
		        p0 = getPoint( cp, r * .76, 90);
		        bg2.draw( new Line2D.Double( p0, p1 ) );
		        
		        // locator points
		        double er = 0.03;
		        p2 = getPoint( p1, r * .2, 0);
		        ell = new Ellipse2D.Double( p2.getX() - r*er/2, p2.getY() -r*er/2, r*er, r*er);
		        bg2.draw(ell);
		        p2 = getPoint( p1, r * .4, 0);
		        ell = new Ellipse2D.Double( p2.getX() - r*er/2, p2.getY() -r*er/2, r*er, r*er);
		        bg2.draw(ell);
		        p2 = getPoint( p1, r * .2, 180);
		        ell = new Ellipse2D.Double( p2.getX() - r*er/2, p2.getY() -r*er/2, r*er, r*er);
		        bg2.draw(ell);
		        p2 = getPoint( p1, r * .4, 180);
		        ell = new Ellipse2D.Double( p2.getX() - r*er/2, p2.getY() -r*er/2, r*er, r*er);
		        bg2.draw(ell);
		        bg2.dispose();
	        } while( backVi.contentsLost() );
    	}
    	g2.drawImage( backVi, 0, 0, this);
    }

    protected synchronized void drawText( final double r, final Point2D cp, Graphics2D g2 ) {
    }

    protected synchronized void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
		Point2D p0, p1, p2, p3, p4, p5, p6, p7, p8, pc;
    	Polygon pol;
    	Ellipse2D ell;
    	Rectangle2D rect;
    	Shape oldClip = g2.getClip();
        Font oldFont = g2.getFont();
        double hir = 1.32;
        
        // draw hi scale
		p1 = new Point2D.Double(cp.getX() - r * .95, cp.getY() - r * 0.92 );
    	rect = new Rectangle2D.Double(p1.getX(), p1.getY(), r * 1.84, r * 1.84 );
    	g2.setClip(rect);

        // bank 0 triangle
		Angle a = new Angle(0);
        p1 = getPoint( cp, -r * .78, 270);
        p2 = getPoint( p1, r * .05, 60);
        p3 = getPoint (p1, r * .05, 120);
        
        pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );

        g2.setColor( clrWhite );
        g2.draw( pol);

		// draw selected heading marker (bottom)
        pc = new Point2D.Double( cp.getX(), cp.getY() + r * .55 );
		a = new Angle( 90 + hi - shdg );

		if( a.getDegrees() < 270 && a.getDegrees() > 132 ) a.setDegrees(132);
		if( a.getDegrees() > 270 || a.getDegrees() < 52 ) a.setDegrees(52);

        p0 = pc; // new Point2D.Double( pc.getX() -r*1.56 + r*3.12/2, pc.getY() -r *1.56 + r*3.12/2);
        p1 = getPoint( p0, r * (hir + .005), a.getDegrees());
		p2 = getPoint( p1, r*.057, a.getDegreesMinus(30));
		p3 = getPoint( p2, r*.025, a.getDegreesMinus(90 ));
		p4 = getPoint( p3, r*.047, a.getDegreesMinus(180));
		p5 = getPoint( p1, r*.057, a.getDegreesPlus(30));
		p6 = getPoint( p5, r*.025, a.getDegreesPlus(90));
		p7 = getPoint( p6, r*.047, a.getDegreesPlus(180));
		pol = new Polygon();
        pol.addPoint( (int)Math.round(p1.getX()), (int)Math.round(p1.getY()) );
        pol.addPoint( (int)Math.round(p2.getX()), (int)Math.round(p2.getY()) );
        pol.addPoint( (int)Math.round(p3.getX()), (int)Math.round(p3.getY()) );
        pol.addPoint( (int)Math.round(p4.getX()), (int)Math.round(p4.getY()) );
        pol.addPoint( (int)Math.round(p7.getX()), (int)Math.round(p7.getY()) );
        pol.addPoint( (int)Math.round(p6.getX()), (int)Math.round(p6.getY()) );
        pol.addPoint( (int)Math.round(p5.getX()), (int)Math.round(p5.getY()) );
		g2.setColor(clrPink);
		g2.draw(pol);
		a = new Angle( 90 + hi - shdg );
		boolean drawLine = a.getDegrees() < 180;
        p1 = getPoint( p0, r * (hir + .005), a.getDegrees());
		if( drawLine ) g2.draw( new Line2D.Double(pc, p1));

        double r1 = r * (hir -.085);
        double majorTickSize = proz(r,6);
        double minorTickSize = proz(r,4.5);
        double majorTickR =  r1 + 4; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/3.0; // minor tick radius
        double minValue = 0;
        double maxValue = 359;
        double minorTickStep = 5;
        double majorTickStep = 10;
        double angleStart = 90;
        double angleExtend = 360;

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( Color.LIGHT_GRAY );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / minorTickStep );
                if( currentAngle.getDegrees() < 35 || currentAngle.getDegrees() > 145 ) 
                	continue;
                p0 = getPoint( pc, minorTickR, currentAngle.getDegrees() );
                p1 = getPoint( pc, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2.7 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0, r * .05 * f);  // minimum 6.0
            g2.setColor( Color.LIGHT_GRAY );
            g2.setStroke( new BasicStroke( (int)(r * 0.0075 )) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / majorTickStep );
                if( currentAngle.getDegrees() < 35 || currentAngle.getDegrees() > 145 ) 
                	continue;

                majorFontSize = Math.max( 6.0, r * (v % 30 == 0 ? .07 : .05) * f);  // minimum 6.0
                Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
                g2.setFont( font);        
                
                p0 = getPoint( pc, majorTickR, currentAngle.getDegrees() );
                p1 = getPoint( pc, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                if( v % 30 == 0 ) {
	            	String s = valueToString( v / 10 );
	
	                TextLayout stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(r * 0.08)), new FontRenderContext(null, true, false));
	            	AffineTransform tat = new AffineTransform();
	            	currentAngle = new Angle( 360 - currentAngle.getDegrees() - 45 );
	            	rect = stl.getBounds();
	            	tat.translate(pc.getX() - rect.getWidth()/2, pc.getY() - stl.getAscent()/2 ); 
	            	tat.rotate( Math.toRadians( currentAngle.getDegrees() ), rect.getCenterX(), rect.getCenterX() );
	            	tat.translate( r *.81, r *.81 );
	            	tat.rotate( Math.toRadians( 135 ), rect.getCenterX(), rect.getCenterX() );
	            	Shape ts = stl.getOutline(tat);
	                g2.fill(ts);
                }
            }
        }	
        ell = new Ellipse2D.Double( pc.getX() -r*hir, pc.getY() -r *hir, r*hir*2, r*hir*2);
        g2.draw(ell);
        
		g2.setClip(oldClip);

    }

    protected synchronized void drawNeedle( final double r, double needleR, final Point2D cp, Graphics2D g2 ) {

    }

    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    	Point2D p1;
    	
		// draw ground speed (top left)
		g2.setColor( clrWhite );
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() - r * .83 );
		drawTextLeft( r, p1, g2, .7, "GS "+Math.round(gs));
		// draw true airspeed
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() - r * .75 );
		drawTextLeft( r, p1, g2, .7, "TAS "+Math.round(tas));
		// draw wind velocity * / direction
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() - r * .67 );
		drawTextLeft( r, p1, g2, .7, ""+Math.round(wd)+"*/"+Math.round(ws));
		
		// heading mag (top center)
		g2.setColor( clrGreen );
		p1 = new Point2D.Double(cp.getX() - r * .15, cp.getY() - r * .83 );
		drawTextRight( r, p1, g2, .8, "HDG");
		p1 = new Point2D.Double(cp.getX() + r * .15, cp.getY() - r * .83 );
		drawTextLeft( r, p1, g2, .8, "MAG");
		g2.setColor( clrWhite );
		p1 = new Point2D.Double(cp.getX(), cp.getY() - r * .83 );
		drawText( r, p1, g2, .8, ""+ Math.round(hi));
		
		// draw nav1 info
		//if( nav1signal != null && (Double)nav1signal.getValue() == -1.0 ) { // nav 1 has signal
		g2.setColor( clrGreen );
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() + r * .67 );
		drawTextLeft( r, p1, g2, .7, "VOR1");
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() + r * .75 );
		drawTextLeft( r, p1, g2, .7, ""+nav1ident.getStringValue());
		p1 = new Point2D.Double(cp.getX() - r * .9, cp.getY() + r * .83 );
		drawTextLeft( r, p1, g2, .7, "DME "+dmeNF.format(navDME1.getDoubleValue()));
		//}

		// draw nav2 info
		//if( nav2signal != null && (Double)nav2signal.getValue() == -1.0 ) { // nav 2 has signal
		g2.setColor( clrGreen );
		p1 = new Point2D.Double(cp.getX() + r * .86, cp.getY() + r * .67 );
		drawTextRight( r, p1, g2, .7, "VOR2");
		p1 = new Point2D.Double(cp.getX() + r * .86, cp.getY() + r * .75 );
		drawTextRight( r, p1, g2, .7, ""+nav2ident.getStringValue());
		p1 = new Point2D.Double(cp.getX() + r * .86, cp.getY() + r * .83 );
		drawTextRight( r, p1, g2, .7, "DME "+dmeNF.format(Math.abs(navDME2.getDoubleValue())));
		//}
    }
}
