package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.AbstractGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;


/**
 * ADF Compass
 * 
 * Dev version 
 * 
 * @author Michael Nagler
 *
 */
public class ADFCompass extends AbstractClockGauge {

	private FSXVariable adfRadial;
	private FSXVariable adfOBS;
	private FSXVariable headingIndicator;
	
	private double hi, vr, obs;

	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );
		setMinValue( 0 );
        setMaxValue( 359 );
        setText("ADF");
        setDecPlaceCount(0);
        setMajorTickStep(30);
        setMinorTickStep(10);
        setAngleExtend(360);
        setAngleStart(90);
        setValue(0);
        //this.index = index;
	    //super.setTextPosition(1.0, 0.45);      
        adfRadial = owner.registerFSXVariable("ADF RADIAL:1");
        adfOBS = owner.registerFSXVariable("ADF CARD:1");
        headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");
	}
	
	public void update() {
		
	 	obs = adfOBS.getDoubleValue();
		vr = adfRadial.getDoubleValue();
		hi = 180 / Math.PI * headingIndicator.getDoubleValue();
		
		repaint();
	}

	protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
       	double r0 = r * .65;
        double r1 = r * .80;

        double majorTickSize = r * .07;
        double minorTickSize = r * .05;
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( minorTickColor );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - obs - hi - minValue) / minorTickStep );
                Point2D p0 = getPoint( cp, minorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - obs - hi - minValue) / majorTickStep );
                Point2D p0 = getPoint( cp, majorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s;
                switch( (int)Math.round( v / 10 )) {
                case  0: s = "N"; break;
                case  9: s = "E"; break;
                case 18: s = "S"; break;
                case 27: s = "W"; break;
                default:
                	 s = valueToString( v / 10 );
                }
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle.getDegrees() ); // marker center point
                
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
            }
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }	
	}		

	protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
		super.drawNeedle(r, needleR, cp, g2, vr);
		// n, w, s, e markers
		Polygon pol;
		Point2D p1, p2, p3;
        double fo=.93;
       	Ellipse2D.Double ello = new Ellipse2D.Double( cp.getX() - r * fo, cp.getY() - r * fo, 2 * r * fo, 2 * r * fo );
        g2.setColor( Color.ORANGE );
		g2.setClip(ello);
        for( double a = 0; a < 360; a += 90 ) {
            pol = new Polygon();
        	p1 = getPoint( cp, r * .85, new Angle( a ).getDegrees() );
        	p2 = getPoint( p1, r * .2, new Angle( a +30 ).getDegrees() );
        	p3 = getPoint( p1, r * .2, new Angle( a -30 ).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        }
    }
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
        // System.out.println("VOR"+index+":"+obs);
        //g2.rotate( Math.PI * (360 - cobs) / 180, cp.getX(), cp.getY() );
   }
}
