package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.RoundRectangle2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;


/**
 * Altimeter Gauge
 * 
 * 
 * @author Michael Nagler
 */
public class Altimeter extends AbstractClockGauge {
	
	private static final long serialVersionUID = 1L;

	private FSXVariable indicatedAltitude;
	private FSXVariable kohlsmanMB;
	private FSXVariable kohlsmanHG;
	
	private double ialt; // indicated altitude value
	private double cialt; // current indicated altitude as shown by the gauge 
	
	public void init( GaugeBook owner, String title ) {
		this.setMinValue( 0 );
	    this.setMaxValue( 1000 );
	    this.setText("Altimeter");
	    this.setDecPlaceCount(0);
	    this.setMajorTickStep(100);
	    this.setMinorTickStep(20);
	    this.setAngleExtend(360);
	    this.setAngleStart(90);
	    setProperty("angleStart", 90 );
	    setProperty("angleExtend", 360 );
	    setProperty("minValue",0.0);
	    setProperty("maxValue",1000.0);
	    setProperty("minorTickStep", 20.0 );
	    setProperty("majorTickStep", 100.0);
	    this.setValue(1021);
	    super.setTextPosition(0, -.25);
	    indicatedAltitude = owner.registerFSXVariable("INDICATED ALTITUDE");
	    kohlsmanMB = owner.registerFSXVariable("KOHLSMAN SETTING MB");
	    kohlsmanHG = owner.registerFSXVariable("KOHLSMAN SETTING HG"); 

		super.init( owner, title );
	}
	
	public void update() {
		ialt = indicatedAltitude.getDoubleValue();
		
		if( Math.round(cialt) != Math.round(ialt)) {
			super.setValue(cialt = ialt);
		}
		repaint();
	}

    protected void conformProperties() {
        // minValue, maxValue 
        double min = Math.min( minValue, maxValue ); 
        double max = Math.max( minValue, maxValue ); 
        
        minValue = min;
        maxValue = max;
        // value - sets value to minvalue if value is outside of range.
        if( value < minValue || value > maxValue ) { 
            //value = (value < minValue ? minValue : maxValue);
        }
        //majorTickStep
        if(majorTickStep >= (maxValue - minValue)){
            majorTickStep = (maxValue - minValue)/10.0;
        }
        // minorTickStep
        if(minorTickStep >= majorTickStep){
            minorTickStep = majorTickStep/4.0;
        }
        
        // angleExtend
        if(angleExtend < 90.0 || angleExtend > 360.0){
            angleExtend = angleExtend < 90.0 ? 90.0 : 360.0;
        }

        if(angleStart < - 180.0 || angleStart > 180.0){
            angleStart = angleStart < -180.0 ? -1800.0 : 180.0;
        }
    }
    
    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
		if( !initdone ) return;
    	super.drawNeedle(r, needleR, cp, g2);
    	double val = ialt / 10;
        double angleCurrent = angleStart + angleExtend * (1.0 - (val - minValue)/(maxValue - minValue));
        Point2D apL = getPoint( cp, needleR*.8, angleCurrent+3 );
        Point2D ap = getPoint( cp, needleR*.9, angleCurrent );
        Point2D apR = getPoint( cp, needleR*.8, angleCurrent-3 );
        
        Angle angleLeft = new Angle( -45 ) ;
        Angle angleRight = new Angle( 45 ) ;

        Point2D pointLeft  = getPoint( cp, needleR * .035, angleLeft.getDegrees() );
        Point2D pointRight = getPoint( cp, needleR * .035, angleRight.getDegrees() );
        
        Polygon pol = new Polygon();
        pol.addPoint( (int)Math.round(pointLeft.getX()), (int)Math.round(pointLeft.getY()) );
        pol.addPoint( (int)Math.round(apL.getX()), (int)Math.round(apL.getY()) );
        pol.addPoint( (int)Math.round(ap.getX()), (int)Math.round(ap.getY()) );
        pol.addPoint( (int)Math.round(apR.getX()), (int)Math.round(apR.getY()) );
        pol.addPoint( (int)Math.round(pointRight.getX()), (int)Math.round(pointRight.getY()) );
        
        g2.setColor( needleColor );
        g2.fill( pol);

        pointLeft  = getPoint( cp, needleR * .12, angleLeft.plus( -180 +20 ));
        pointRight = getPoint( cp, needleR * .12, angleLeft.plus( -90 -20 ));

        apL = getPoint( cp, needleR * .12, angleLeft.plus( 20 ));
        apR = getPoint( cp, needleR * .12, angleLeft.plus( 90 - 20 ));

        pol = new Polygon();

        pol.addPoint( (int)Math.round(pointLeft.getX()), (int)Math.round(pointLeft.getY()) );
        pol.addPoint( (int)Math.round(pointRight.getX()), (int)Math.round(pointRight.getY()) );
        pol.addPoint( (int)Math.round(apL.getX()), (int)Math.round(apL.getY()) );
        pol.addPoint( (int)Math.round(apR.getX()), (int)Math.round(apR.getY()) );
        
        g2.setColor( Color.GRAY );
        g2.fill( pol);
        
    }
    
    protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
		if( !initdone ) return;
        double r0 = proz(r,65);
        double r1 = proz(r,80);

        double majorTickSize = proz(r,7);
        double minorTickSize = proz(r,5);
        
        double majorTickR =  r1 + 4; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
            g2.setColor( minorTickColor );
            g2.setStroke( new BasicStroke(1.f) );            
            double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            for(double i= angleStart + angleExtend; i >= angleStart ; i-=minorGradSteps){
                Point2D p0 = getPoint( cp, minorTickR, i );
                Point2D p1 = getPoint( cp, minorTickR+minorTickSize, i );
                g2.draw( new Line2D.Double(p0,p1) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 7.0,proz(r,14.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v< maxValue; v+=majorTickStep ) {
                double currentAngle = angleStart + angleExtend - stepAngle * (v - minValue) / majorTickStep;
                Point2D p0 = getPoint( cp, majorTickR, currentAngle );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s = valueToString( Math.round( v / 100 ));
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle ); // marker center point
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
                
            }
            
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }
    }
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
		if( !initdone ) return;
        //draw current value
    	if( kohlsmanMB == null ) return;
    	
        g2.setColor( valueBackgroundColor  );
        String s = valueToString( kohlsmanMB.getDoubleValue() );
        // center point of current value area
        double rv = r * .45;  
        Point2D p = getPoint( cp, rv, 0 ); //  angleStart + angleExtend + (360 - angleExtend)/2 ); // value center point        
        double h = r * .16;   // height of current value area 
        g2.setFont( g2.getFont().deriveFont( Font.BOLD, (float)(h * 0.7)  ) );
        
        Dimension size = getStringSize( s, g2 );
        int l = s.length();
        double f = l < 3 ? 2.0 : Math.max(2.0 - l*0.1, 1.1);
        double w  = size.getWidth() * f; // width of current value area 
        double corner = Math.min( 10, Math.max( h, w ) / 4.0); // maximum 10
        
        RoundRectangle2D cvArea = new RoundRectangle2D.Double(p.getX()- w/2,p.getY()-h/2,w,h,corner,corner);
        g2.fill( cvArea );
        
        Stroke old = g2.getStroke();
        g2.setColor( scaleLineColor );
        g2.setStroke( new BasicStroke(h < 20 ? 1f : (h < 40 ? 2f : 3f)) );
        g2.draw( cvArea );
        
        g2.setStroke( old );
        
		g2.setColor( valueColor );
        
        double x0 = cvArea.getCenterX() - size.getWidth()/2.0;
        double y0 = cvArea.getCenterY() + size.getHeight()/2.0;
        g2.drawString( s,(float)x0,(float)y0);
        
        // mb
        g2.setColor( textColor );
        super.drawText(r, cp, g2, new Point2D.Double(.45,.25), 1, "mb");
    }   
}
