package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;


/**
 * VOR Compass
 * 
 * @author Michael Nagler
 *
 */
public class CourseDeviationIndicator extends AbstractClockGauge {

	protected FSXVariable vorRadial, vorOBS, vorCDI, vorBCF, vorToFrom;
	
	protected int index;
	protected double vr, obs, cdi, bcf, tf;

	public void init( GaugeBook owner, String title, int index ) {
		super.init( owner, title );
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setText(" ");
        this.setDecPlaceCount(0);
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);
        this.setAngleExtend(360);
        this.setAngleStart(90);
        this.setValue(0);
        this.index = index;

        vorOBS = owner.registerFSXVariable("NAV OBS:"+index );
        vorCDI = owner.registerFSXVariable("NAV CDI:"+index );
        vorToFrom = owner.registerFSXVariable("NAV TOFROM:"+index ); // Nav TO/FROM flag: 0 = Off, 1 = TO, 2 = FROM
	}
	
	public void init( GaugeBook owner, String title ) {
		init( owner, title, 2 );
	}
	
	public void update() {
		if( obs != Math.round(vorOBS.getDoubleValue())
		 || cdi != vorCDI.getDoubleValue()
		 || tf  != vorToFrom.getDoubleValue() ) {

			obs = vorOBS.getDoubleValue();
			cdi = vorCDI.getDoubleValue();
			tf = vorToFrom.getDoubleValue();
		}
		repaint();
	}

	protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
       	double r0 = r * .65;
        double r1 = r * .80;

        double majorTickSize = r * .07;
        double minorTickSize = r * .05;
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( minorTickColor );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - obs - minValue) / minorTickStep );
                Point2D p0 = getPoint( cp, minorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - obs - minValue) / majorTickStep );
                Point2D p0 = getPoint( cp, majorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s;
                switch( (int)Math.round( v / 10 )) {
                case  0: s = "N"; break;
                case  9: s = "E"; break;
                case 18: s = "S"; break;
                case 27: s = "W"; break;
                default:
                	 s = valueToString( v / 10 );
                }
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle.getDegrees() ); // marker center point
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );

/*                
                TextLayout stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(r * 0.135)), new FontRenderContext(null, true, false));
            	AffineTransform tat = new AffineTransform();
            	currentAngle = new Angle( 360 - currentAngle.getDegrees() - 135);
            	Rectangle2D rect = stl.getBounds();
            	tat.translate(cp.getX() - rect.getWidth()/2, cp.getY() - stl.getAscent()/2 ); 
            	tat.rotate( Math.toRadians( currentAngle.getDegrees() ), rect.getCenterX(), rect.getCenterX() );
            	tat.translate( r *.45, r *.45 );
            	tat.rotate( Math.toRadians( 135 ), rect.getCenterX(), rect.getCenterX() );
            	Shape ts = stl.getOutline(tat);
                g2.fill(ts);
*/
            }
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }	
	}		

	protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
		this.setMinValue( -128 );
        this.setMaxValue(  128 );
        this.setAngleExtend(70);
        this.setAngleStart(235);

        // cdi needle
        Point2D mp = getPoint( cp, r*.6, 90 );
        double tickRate = 25.6;
        double myValue = -1 * cdi;
        if( myValue < minValue || myValue > maxValue ) { 
        	myValue = (myValue < minValue ? minValue : maxValue);
        }
        double angleCurrent = angleStart + angleExtend * (1.0 - (myValue - minValue)/(maxValue - minValue));

        Point2D ap = getPoint( mp, needleR * 1.5, angleCurrent );
        
        g2.setColor( Color.WHITE );
        g2.setStroke( new BasicStroke(2f));
        g2.draw( new Line2D.Double(mp, ap));
        
    	g2.setColor( minorTickColor );
        g2.setStroke( new BasicStroke(2f));
        double stepAngle = angleExtend * tickRate/(maxValue-minValue);
        for( double v = minValue; v<= maxValue; v+=tickRate) {
            Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - minValue) / tickRate );
            Point2D p0 = getPoint( mp, r * 0.9, currentAngle.getDegrees() );
            Point2D p1 = getPoint( mp, r * 0.905, currentAngle.getDegrees() );
            g2.draw( new Line2D.Double( p0, p1 ) );
        }

        // draw to / from triangle
        // 0 = Off, 1 = TO, 2 = FROM
        Polygon pol;
        Point2D p1, p2, p3, p4;
        Shape oldClip = g2.getClip();
        switch( (int)tf ) {
        case 1: // to
            pol = new Polygon();
        	p1 = getPoint(cp, r*.4, 7);
        	p2 = getPoint(cp, r*.3, -7);
        	p3 = new Point2D.Double( p2.getX() + r*.2, p2.getY());
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        	break;
        case 2: // from
            pol = new Polygon();
        	p1 = getPoint(cp, r*.4, -7);
        	p2 = getPoint(cp, r*.3, 7);
        	p3 = new Point2D.Double( p2.getX() + r*.2, p2.getY());
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        	break;
        default: // off or error
        	Rectangle2D rect;
	    	p1 = getPoint(cp, r*.3, 7);
	    	p2 = getPoint(cp, r*.5, -7);
	    	rect = new Rectangle2D.Double(p1.getX(),p1.getY(),p2.getX()-p1.getX(),p2.getY()-p1.getY());
	    	g2.fill(rect.getBounds2D());
	    	g2.clip(rect.getBounds2D());
	    	g2.setColor(Color.RED);
	    	g2.setStroke(new BasicStroke((float)(r*.04)));
	    	p1 = getPoint(cp, r*.23, -18);
	    	p2 = getPoint(p1, r*.15, 45);
	    	for( int i = 0; i < 3; i ++ ) {
		    	g2.draw(new Line2D.Double(p1, p2));
	    		p1.setLocation( p1.getX()+r*.12, p1.getY());
	    		p2.setLocation( p2.getX()+r*.12, p2.getY());
	        }
        }
        g2.setClip(oldClip);
        
        this.setAngleExtend(360);
        this.setAngleStart(270);
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);

		// n, w, s, e markers
        double fo=.93;
       	Ellipse2D.Double ello = new Ellipse2D.Double( cp.getX() - r * fo, cp.getY() - r * fo, 2 * r * fo, 2 * r * fo );
        g2.setColor( Color.ORANGE );
		g2.setClip(ello);
        for( double a = 0; a < 360; a += 90 ) {
            pol = new Polygon();
        	p1 = getPoint( cp, r * .85, new Angle( a ).getDegrees() );
        	p2 = getPoint( p1, r * .2, new Angle( a +30 ).getDegrees() );
        	p3 = getPoint( p1, r * .2, new Angle( a -30 ).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        }
        
	}
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
        // System.out.println("VOR"+index+":"+obs);
        //g2.rotate( Math.PI * (360 - cobs) / 180, cp.getX(), cp.getY() );
    }
    

}
