package org.gaugebook.gauges.aviation.simple;

import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Point2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;

/**
 * EGT Fuel Flow Gauge
 * 
 * @author Michael Nagler
 */
public class EGTFuelFlow extends AbstractClockGauge {
	
	private FSXVariable varEGT, varFuelFlow;

	private double egt = 0, cegt = 0, ff = 0, cff = 0;
	private final int angleStartLeft = 315, angleExtendLeft = 90, angleStartRight = -225, angleExtendRight = 90;
	private final int minValueLeft = 0, maxValueLeft = 2000, minValueRight = 0, maxValueRight = 20;
	private final int minorTickStepLeft = 250, majorTickStepLeft = 500, minorTickStepRight = 1, majorTickStepRight = 5;
	
	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );
		this.setMinValue( 0 );
        this.setMaxValue( 30 );
        this.setText("");
        this.setDecPlaceCount(0);
        this.setMajorTickStep(5);
        this.setMinorTickStep(1);
        this.setAngleExtend(angleExtendLeft);
        this.setAngleStart(angleStartLeft);
        this.setValue(0);
        this.setShowValue(false);
	    super.setTextPosition(1.0, 0.2);
	    super.setTextHeight( 1.4 ); 
	    varEGT  = owner.registerFSXVariable("ENG EXHAUST GAS TEMPERATURE:1");
	    varFuelFlow = owner.registerFSXVariable("ENG FUEL FLOW GPH:1");
    }
	
	public void update() {
		egt = varEGT.getDoubleValue();
		ff = varFuelFlow.getDoubleValue();
		if( Math.round(cegt) != Math.round(egt) || Math.round(cff) != Math.round(ff)) {
			setValue(cegt = egt );
			cff = ff;
		}
		
		repaint();
	}
	
    protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
    	double f = 1;
    	Point2D left = new Point2D.Double(cp.getX() - r * f, cp.getY());
    	Point2D right= new Point2D.Double(cp.getX() + r * f, cp.getY());
        setAngleExtend(angleExtendRight);
        setAngleStart(angleStartRight);
        setMinorTickStep(minorTickStepRight);
        setMajorTickStep(majorTickStepRight);
        setMinValue(minValueRight);
        setMaxValue(maxValueRight);
    	super.drawScale(r, right, g2, Direction.CLOCKWISE);
        setAngleExtend(angleExtendLeft);
        setAngleStart(angleStartLeft);
        setMinorTickStep(minorTickStepLeft);
        setMajorTickStep(majorTickStepLeft);
        setMinValue(minValueLeft);
        setMaxValue(maxValueLeft);
    	super.drawScale(r, left, g2, Direction.ANTICLOCKWISE);
    	
    }

    protected void drawNeedle( final double r, double needleR, final Point2D cp, Graphics2D g2 ) {
    	Shape oldClip = g2.getClip();
    	double cf = .87; // clip factor
    	Ellipse2D.Double ell = new Ellipse2D.Double( cp.getX() - r * cf, cp.getY() - r * cf, 2 * r * cf, 2 * r * cf );
		g2.setClip(ell);

    	double f = 1;
    	Point2D left = new Point2D.Double(cp.getX() - r * f, cp.getY());
    	Point2D right= new Point2D.Double(cp.getX() + r * f, cp.getY());
        setAngleExtend(angleExtendRight);
        setAngleStart(angleStartRight);
        setMinorTickStep(minorTickStepRight);
        setMajorTickStep(majorTickStepRight);
        setMinValue(minValueRight);
        setMaxValue(maxValueRight);
        double v = cff;
        if( v < minValueRight || v > maxValueRight ) { 
            v = (v < minValueRight ? minValueRight : maxValueRight);
        }
    	super.drawNeedle(r, r * .7, right, g2, v);
        setAngleExtend(angleExtendLeft);
        setAngleStart(angleStartLeft);
        setMinorTickStep(minorTickStepLeft);
        setMajorTickStep(majorTickStepLeft);
        setMinValue(minValueLeft);
        setMaxValue(maxValueLeft);
        v = cegt - 459.67; // rankine in F
        if( v < minValueLeft || v > maxValueLeft ) { 
            v = (v < minValueLeft ? minValueLeft : maxValueLeft);
        }
    	super.drawNeedle(r, r * .7, left, g2, v, Direction.ANTICLOCKWISE ); 
    	
		g2.setClip(oldClip);
    }
    
    protected void drawText( final double r, final Point2D cp, Graphics2D g2 ) {
    	super.drawText(r, cp, g2);
    	super.drawText(r, cp, g2, new Point2D.Double(-.65, .1), 1.2, "EGT");
    	super.drawText(r, cp, g2, new Point2D.Double( .65, .05), 1.1, "Fuel");
    	super.drawText(r, cp, g2, new Point2D.Double( .65, .15), 1.1, "Flow");
    	
    }

}
