package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;


/**
 * Heading Indicator
 * 
 * @author Michael Nagler
 *
 */
public class HeadingIndicator extends AbstractClockGauge {

	private FSXVariable headingIndicator;
	private FSXVariable headingLockDir;
	
	private double hi, hld;

	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setText(""); // Course Indicator
        this.setDecPlaceCount(0);
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);
        this.setAngleExtend(359);
        this.setAngleStart(90);
        this.setValue(0);
        this.setNeedleColor(Color.WHITE);
      
        headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");
        headingLockDir = owner.registerFSXVariable("AUTOPILOT HEADING LOCK DIR");
	}
	
	public void update() {
		hi = ( 180 / StrictMath.PI ) * headingIndicator.getDoubleValue();
		hld = headingLockDir.getDoubleValue();
		repaint();
	}

	protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
       	double r0 = proz(r,65);
        double r1 = proz(r,80);

        double majorTickSize = proz(r,7);
        double minorTickSize = proz(r,5);
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( minorTickColor );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / minorTickStep );
                Point2D p0 = getPoint( cp, minorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / majorTickStep );
                Point2D p0 = getPoint( cp, majorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s;
                switch( (int)Math.round( v / 10 )) {
                case  0: s = "N"; break;
                case  9: s = "E"; break;
                case 18: s = "S"; break;
                case 27: s = "W"; break;
                default:
                	 s = valueToString( v / 10 );
                }
                /*
                TextLayout stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(r * 0.135)), new FontRenderContext(null, true, false));
            	AffineTransform tat = new AffineTransform();
            	currentAngle = new Angle( 360 - currentAngle.getDegrees() - 135);
            	Rectangle2D rect = stl.getBounds();
            	tat.translate(cp.getX() - rect.getWidth()/2, cp.getY() - stl.getAscent()/2 ); 
            	tat.rotate( Math.toRadians( currentAngle.getDegrees() ), rect.getCenterX(), rect.getCenterX() );
            	tat.translate( r *.42, r *.42 );
            	tat.rotate( Math.toRadians( 135 ), rect.getCenterX(), rect.getCenterX() );
            	Shape ts = stl.getOutline(tat);
                g2.fill(ts);
                */
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle.getDegrees() ); // marker center point
                
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
            }
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }	
	}		

    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    	super.drawNeedle(r, needleR*.8, cp, g2, 0); // chi);
    	super.drawNeedle(r, needleR*.7, cp, g2, 180); // 360 +chi - 180);
    	
    	// heading marker
        Angle angleCurrent = new Angle( angleStart + angleExtend * (hld - hi - minValue)/(maxValue - minValue));
        angleCurrent = new Angle( 180 - angleCurrent.getDegrees());
        
        Point2D ap = getPoint( cp, r * .8, angleCurrent.getDegrees() );
        
        Angle angleLeft = new Angle( angleCurrent.getDegrees() + 5 );
        Angle angleRight = new Angle( angleCurrent.getDegrees() - 5 );

        Point2D pointLeft  = getPoint( cp, r * .92, angleLeft.getDegrees()  );
        Point2D pointRight = getPoint( cp, r * .92, angleRight.getDegrees() );
        
        Polygon pol = new Polygon();
        pol.addPoint( (int)Math.round(pointLeft.getX()), (int)Math.round(pointLeft.getY()) );
        pol.addPoint( (int)Math.round(ap.getX()), (int)Math.round(ap.getY()) );
        pol.addPoint( (int)Math.round(pointRight.getX()), (int)Math.round(pointRight.getY()) );
        
        g2.setColor( new Color( 224, 142, 48 ));
        g2.fill( pol);
        
        // n, w, s, e markers
        for( double a = 0; a < 360; a += 90 ) {
        	pointLeft  = getPoint( cp, r * .86, a );
        	pointRight = getPoint( cp, r * .92, a );
        	g2.setStroke(new BasicStroke(4f));
        	g2.draw( new Line2D.Double( pointLeft, pointRight ));
        }
    }
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    }
}
