package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.Polygon;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.Area;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.VolatileImage;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;
import org.gaugebook.math.Trigonometry;


/**
 * Heading Indicator
 * 
 * @author Michael Nagler
 *
 */
public class HorizontalSituationIndicator extends AbstractClockGauge {

	private static final long serialVersionUID = 1L;

	private FSXVariable headingIndicator, headingLockDir;
	private FSXVariable vorOBS, vorCDI, vorToFrom, vorGSAngle, vorGSFlag;
	
	private double hi, hld, obs, cdi, tf, gsa, gsf;

	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setText(""); // Course Indicator
        this.setDecPlaceCount(0);
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);
        this.setAngleExtend(359);
        this.setAngleStart(90);
        this.setValue(0);
        this.setNeedleColor(Color.WHITE);
	    //super.setTextPosition(1.1, 0.7);
      
        headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");
        headingLockDir = owner.registerFSXVariable("AUTOPILOT HEADING LOCK DIR");
        vorOBS = owner.registerFSXVariable("NAV OBS:1");
        vorCDI = owner.registerFSXVariable("NAV CDI:1");
        vorToFrom = owner.registerFSXVariable("NAV TOFROM:1");
		vorGSAngle = owner.registerFSXVariable("NAV GLIDE SLOPE ERROR:1");
		vorGSFlag = owner.registerFSXVariable("NAV GS FLAG:1");
	}

	int lastWidth = 0, lastHeight = 0;
	
    /**
     * {@inheritDoc}
     */
    @Override
    public void paint( Graphics g ) {
        int w = getWidth();
        int h = getHeight();
        GraphicsConfiguration gc = getGraphicsConfiguration();
        if( update || w != lastWidth || h != lastHeight || vi != null && vi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE) {
            update = false;
            lastWidth = w;
            lastHeight = h;
            conformProperties();

    		vi = gc.createCompatibleVolatileImage( getWidth(), getHeight() );
            
	        do {
	        	gc = this.getGraphicsConfiguration();
	        	if( this.vi.validate(gc) == VolatileImage.IMAGE_INCOMPATIBLE ) {
	        		vi = gc.createCompatibleVolatileImage( getWidth(), getHeight() );
	        	}
	        	
	        	Graphics2D g2 = vi.createGraphics();
	            g2.setRenderingHint( RenderingHints.KEY_ANTIALIASING, (antialias ? RenderingHints.VALUE_ANTIALIAS_ON : RenderingHints.VALUE_ANTIALIAS_OFF ));

	            double r =  Math.min( w, h ) / 2 - 5;            // radius
	            Point2D cp = new Point2D.Double( w / 2, h / 2 ); // center point
	            drawBackground( r, cp, g2 );
	            drawText(r,cp,g2);
	            double r0 = r * .55;
	            double r1 = r * .58;
	            double majorTickSize = r * .04;
	            double minorTickSize = r * .02;
	            double majorTickR =  r1 + 4; // major tick radius 
	            double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
	            double needleR = majorTickR + 1.5 * majorTickSize;
	            drawValue(r,needleR, cp,g2);
	            drawScale(r,cp,g2);
	            drawNeedle( r, needleR, cp, g2 );
	            g2.dispose();
	        } while( vi.contentsLost() );
        }
        g.drawImage( vi, 0, 0, this );
    }
    
	public void update() {
		hi  = ( 180 / StrictMath.PI ) * headingIndicator.getDoubleValue();
		hld = headingLockDir.getDoubleValue();
		obs = vorOBS.getDoubleValue();
		cdi = vorCDI.getDoubleValue();
		tf  = vorToFrom.getDoubleValue();
	 	gsa = vorGSAngle.getDoubleValue();
		gsf = vorGSFlag.getDoubleValue();

		repaint();
	}

	protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
       	double r0 = r * .65;
        double r1 = r * .80;

        double majorTickSize = r * .07;
        double minorTickSize = r * .05;
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( minorTickColor );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v - hi - minValue) / minorTickStep );
                Point2D p0 = getPoint( cp, minorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend -180 - stepAngle * (v - hi - minValue) / majorTickStep );
                Point2D p0 = getPoint( cp, majorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s;
                switch( (int)Math.round( v / 10 )) {
                case  0: s = "N"; break;
                case  9: s = "E"; break;
                case 18: s = "S"; break;
                case 27: s = "W"; break;
                default:
                	 s = valueToString( v / 10 );
                }
                /*
                TextLayout stl = new TextLayout( s, new Font("Helvetica", 1, (int)Math.round(r * 0.135)), new FontRenderContext(null, true, false));
            	AffineTransform tat = new AffineTransform();
            	currentAngle = new Angle( 360 - currentAngle.getDegrees() - 135);
            	Rectangle2D rect = stl.getBounds();
            	tat.translate(cp.getX() - rect.getWidth()/2, cp.getY() - stl.getAscent()/2 ); 
            	tat.rotate( Math.toRadians( currentAngle.getDegrees() ), rect.getCenterX(), rect.getCenterX() );
            	tat.translate( r *.42, r *.42 );
            	tat.rotate( Math.toRadians( 135 ), rect.getCenterX(), rect.getCenterX() );
            	Shape ts = stl.getOutline(tat);
                g2.fill(ts);
                */
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle.getDegrees() ); // marker center point
                
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
            }
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }	
	}		

    protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    	Angle angleCurrent;
    	Polygon pol;
        Point2D p1, p2, p3, p4;
    	Shape oldClip = g2.getClip();
        double fo = .93, fi = .40, tickRate;
    	Ellipse2D.Double ello = new Ellipse2D.Double( cp.getX() - r * fo, cp.getY() - r * fo, 2 * r * fo, 2 * r * fo );
    	Ellipse2D.Double elli = new Ellipse2D.Double( cp.getX() - r * fi, cp.getY() - r * fi, 2 * r * fi, 2 * r * fi );
 
    	// gs flag
        // drawText( final double r, final Point2D cp, Graphics2D g2, Point2D position, double textHeight, String text )
        super.drawText(r, cp, g2, new Point2D.Double(0.15,-0.015),1,"GS");
    	p1 = getPoint(cp, r*.28, 30);
    	p2 = getPoint(p1, r*.20, -26);
    	g2.setColor(Color.WHITE);
    	Rectangle2D rect = new Rectangle2D.Double(p1.getX(),p1.getY(),p2.getX()-p1.getX(),p2.getY()-p1.getY());
    	g2.fill(rect);
        if( gsf == 1.0 ) {
        	// draw ON
        	g2.setColor(Color.BLACK);
            super.drawText(r, cp, g2, new Point2D.Double(0.335,-0.025),0.9,"ON");
        }
        else {
        	g2.fill(rect.getBounds2D());
        	g2.clip(rect.getBounds2D());
        	g2.setColor(Color.RED);
        	g2.setStroke(new BasicStroke((float)(r*.04)));
        	p1 = getPoint(cp, r*.16, 16);
        	p2 = getPoint(p1, r*.15, 45);
        	for( int i = 0; i < 3; i ++ ) {
    	    	g2.draw(new Line2D.Double(p1, p2));
        		p1.setLocation( p1.getX()+r*.12, p1.getY());
        		p2.setLocation( p2.getX()+r*.12, p2.getY());
            }
            g2.setClip(oldClip);
        }
        
    	// draw to / from triangle 
        // 0 = Off, 1 = TO, 2 = FROM
        oldClip = g2.getClip();
        g2.setColor(Color.WHITE);
        angleCurrent = new Angle( angleStart + angleExtend * (obs - hi - minValue)/(maxValue - minValue));
        angleCurrent.invert().add(180);
        switch( (int)tf ) {
        case 1: // to
            pol = new Polygon();
        	p1 = getPoint(cp, r*.3 , new Angle( angleCurrent.getDegrees() +90 ).getDegrees() );
        	p2 = getPoint(p1, r*.15, new Angle( angleCurrent.getDegrees() -45 ).getDegrees() );
        	p3 = getPoint(p1, r*.15, new Angle( angleCurrent.getDegrees() -135).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        	break;
        case 2: // from
            pol = new Polygon();
        	p1 = getPoint(cp, r*.3 , new Angle( angleCurrent.getDegrees() +270).getDegrees() );
        	p2 = getPoint(p1, r*.15, new Angle( angleCurrent.getDegrees() +135).getDegrees() );
        	p3 = getPoint(p1, r*.15, new Angle( angleCurrent.getDegrees() +45 ).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        	break;
        default: // off or error
	    	p1 = getPoint(cp, r* 1.1, 125);
	    	p2 = getPoint(p1, r*.5, 330);
	    	RoundRectangle2D rrect = new RoundRectangle2D.Double(p1.getX(),p1.getY(),p2.getX()-p1.getX(),p2.getY()-p1.getY(), r*.1, r*.1);
	    	g2.clip(ello);
	    	g2.setColor(Color.RED);
	    	g2.fill(rrect);
	    	Color oc = textColor;
	    	textColor = Color.BLACK;
	        super.drawText(r, cp, g2, new Point2D.Double(-0.34,-0.67),1,"NAV");
	        textColor = oc;
        }
        g2.setClip(oldClip);
        
        this.setAngleExtend(360);
        this.setAngleStart(270);
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);

        g2.setClip(oldClip);

        // gs needle
        this.setAngleExtend(50);
        this.setAngleStart(-25);
		this.setMinValue( -1 );
        this.setMaxValue(  1 );

        p1 = getPoint( cp, r*.6, 180 );
        tickRate = .5;
        double val = gsa;
        if( val < minValue || val > maxValue ) { 
        	val = (val < minValue ? minValue : maxValue);
        }
        angleCurrent = new Angle( angleStart + angleExtend * (1.0 - (val - minValue)/(maxValue - minValue)));

        p2 = Trigonometry.getPoint( p1, needleR * 1.7, angleCurrent );
        
        g2.setColor( Color.WHITE );
        g2.setStroke( new BasicStroke((float)(r*.01)));
        g2.draw( new Line2D.Double(p1, p2));
        
        // gs scale
        g2.setColor( minorTickColor );
        g2.setStroke( new BasicStroke((float)(r*.02)));
        double stepAngle = angleExtend * tickRate/(maxValue-minValue);

        for( double v = minValue; v<= maxValue; v+=tickRate) {
            angleCurrent = new Angle( angleStart + angleExtend - stepAngle * (v - minValue) / tickRate );
            p3 = Trigonometry.getPoint( p1, r*1, angleCurrent );
            p4 = Trigonometry.getPoint( p1, r*1.04, angleCurrent );
            g2.draw( new Line2D.Double( p3, p4 ) );
        }

        g2.setClip(oldClip);
        
        // reset values
        this.setAngleExtend(360);
        this.setAngleStart(270);
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);

    	// heading marker
    	g2.setClip(ello);
        angleCurrent = new Angle( angleStart + angleExtend * (hld - hi - minValue)/(maxValue - minValue));
        angleCurrent = new Angle( 270 - angleCurrent.getDegrees());

        p1 = getPoint( cp, r * .85, angleCurrent.getDegrees() );
        p2 = getPoint( cp, r * .70, angleCurrent.add(5).getDegrees());
        p3 = getPoint( cp, r * .70, angleCurrent.sub(10).getDegrees());
        
        pol = new Polygon();
    	pol.addPoint((int)p1.getX(),(int)p1.getY());
    	pol.addPoint((int)p2.getX(),(int)p2.getY());
    	pol.addPoint((int)p3.getX(),(int)p3.getY());
        
        g2.setColor( new Color( 224, 142, 48 ));
    	g2.setStroke(new BasicStroke((float)(r*.02)));
        g2.draw( pol);
        
        // cdi needle
        // outer part (top and bottom) / todo: setClip
    	Area area = new Area( ello );
    	area.subtract( new Area( elli ));
    	g2.setClip(area);
    	
        tickRate = 25.6;
        angleCurrent = new Angle( angleStart + angleExtend * (obs - hi - minValue)/(maxValue - minValue));
        angleCurrent = angleCurrent.invert();

        p1 = getPoint( cp, r*.7, angleCurrent.getDegrees() );
        p2 = getPoint( p1, r * 1.3, angleCurrent.add(180).getDegrees() );
        
        g2.setColor( Color.YELLOW );
        g2.setStroke( new BasicStroke((float)(r*.02)));
        g2.draw( new Line2D.Double(p1, p2));
        
        // inner part / todo
    	g2.setClip(oldClip);
        p1 = getPoint( cp, r * .4 * cdi / 128, angleCurrent.add(90).getDegrees() );
        // inner scale
        g2.setColor( Color.WHITE );
        g2.setStroke( new BasicStroke((float)(r*.03)));
        for( double i = -.4; i <= .4; i += .2 ) {
            p2 = getPoint( cp, r * i, angleCurrent.getDegrees() );
            g2.draw( new Line2D.Double(p2, p2));
        }
        
        // inner line
        p2 = getPoint( p1, r * .4, angleCurrent.add(90).getDegrees() );
        p3 = getPoint( p1, r * .4, angleCurrent.add(180).getDegrees() );
        g2.setColor( Color.YELLOW );
        g2.setStroke( new BasicStroke((float)(r*.02)));
        g2.draw( new Line2D.Double(p2, p3));
        // angleCurrent.add(90);
        
        // spitze
        p1 = getPoint( cp, r*.75, angleCurrent.add(180).getDegrees() );
        p2 = getPoint( cp, r * .6, angleCurrent.add(5).getDegrees());
        p3 = getPoint( cp, r * .6, angleCurrent.sub(10).getDegrees());
        
        pol = new Polygon();
    	pol.addPoint((int)p1.getX(),(int)p1.getY());
    	pol.addPoint((int)p2.getX(),(int)p2.getY());
    	pol.addPoint((int)p3.getX(),(int)p3.getY());
        
        g2.fill( pol);

		// n, w, s, e markers
		g2.setColor( Color.ORANGE );
		g2.setClip(ello);
        for( double a = 0; a < 360; a += 90 ) {
            pol = new Polygon();
        	p1 = getPoint( cp, r * .85, new Angle( a ).getDegrees() );
        	p2 = getPoint( p1, r * .2, new Angle( a +30 ).getDegrees() );
        	p3 = getPoint( p1, r * .2, new Angle( a -30 ).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        }
        
    	// line plane
        g2.setStroke( new BasicStroke((float)(r*.02)));
        p1 = getPoint(cp, r*.1, 180);
        p2 = getPoint(cp, r*.1, 0);
        g2.draw( new Line2D.Double(p1,p2));
        p1 = getPoint(cp, r*.1, 90);
        p2 = getPoint(cp, r*.15, 270);
        g2.draw( new Line2D.Double(p1,p2));
        p1 = getPoint(cp, r*.125, 250);
        p2 = getPoint(cp, r*.125, 290);
        g2.draw( new Line2D.Double(p1,p2));
    }
    
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
    	// System.out.println("MH:"+cmh);
        // g2.rotate( Math.PI * (360 - chi) / 180, cp.getX(), cp.getY() );
    }
}
