package org.gaugebook.gauges.aviation.simple;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Stroke;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import org.gaugebook.AbstractClockGauge;
import org.gaugebook.FSXVariable;
import org.gaugebook.GaugeBook;
import org.gaugebook.math.Angle;


/**
 * Radio Magnetic Indicator (RMI)
 * 
 * @author Michael Nagler
 */
public class RadioMagneticIndicator extends AbstractClockGauge {

	private FSXVariable vor2Radial, adfRadial, headingIndicator;
	
	private double ar, vr, hi;

	@Override
	public void init( GaugeBook owner, String title ) {
		super.init( owner, title );
		this.setMinValue( 0 );
        this.setMaxValue( 359 );
        this.setText("ADF");
        this.setDecPlaceCount(0);
        this.setMajorTickStep(30);
        this.setMinorTickStep(10);
        this.setAngleExtend(360);
        this.setAngleStart(90);
        this.setValue(0);
        vor2Radial = owner.registerFSXVariable("NAV RADIAL:2");
        adfRadial = owner.registerFSXVariable("ADF RADIAL:1");
        headingIndicator = owner.registerFSXVariable("HEADING INDICATOR");

	}
	
	public void update() {
		ar = adfRadial.getDoubleValue();
		vr = vor2Radial.getDoubleValue();
		hi = 180 / Math.PI * headingIndicator.getDoubleValue();
		repaint();
	}

	@Override
	protected void drawScale( final double r, final Point2D cp, Graphics2D g2 ) {
       	double r0 = r * .65;
        double r1 = r * .80;

        double majorTickSize = r * .07;
        double minorTickSize = r * .05;
        
        double majorTickR =  r1 + 5; // major tick radius 
        double minorTickR =  majorTickR + (majorTickSize - minorTickSize)/2.0; // minor tick radius
        
        Stroke oldStroke = g2.getStroke();
        Font   oldFont   = g2.getFont();

        // draw minor ticks
        if(minorTickStep > 0){
        	g2.setColor( minorTickColor );
        	g2.setStroke( new BasicStroke(1.f) );            
            // double minorGradSteps = angleExtend * minorTickStep / (maxValue - minValue);
            double stepAngle = angleExtend * minorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=minorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v -hi - minValue) / minorTickStep );
                Point2D p0 = getPoint( cp, minorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, minorTickR + minorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
            }
        }
        
        // draw major ticks and markers
        if(majorTickStep > 0){
            double markerR = majorTickR + 2 * majorTickSize;
            double l = valueToString( maxValue ).length();
            if(l > 3 && decPlaceCount > 0){
                l--;
            }
            double f = l < 4 ? 1.0 : 1.0 - 0.1 * (l - 3);   // font factor  
            double majorFontSize = Math.max( 6.0,proz(r,12.0) * f);  // minimum 6.0
            Font font = oldFont.deriveFont(majorFontSize < 8.0 ? Font.PLAIN : Font.BOLD, (float)majorFontSize );
            g2.setFont( font);        
            g2.setColor( majorTickColor );
            g2.setStroke( new BasicStroke(2.f) );            
            
            double stepAngle = angleExtend * majorTickStep/(maxValue-minValue);
            for( double v = minValue; v<= maxValue; v+=majorTickStep ) {
                Angle currentAngle = new Angle( angleStart + angleExtend - stepAngle * (v -hi - minValue) / majorTickStep );
                Point2D p0 = getPoint( cp, majorTickR, currentAngle.getDegrees() );
                Point2D p1 = getPoint( cp, majorTickR + majorTickSize, currentAngle.getDegrees() );
                g2.draw( new Line2D.Double( p0, p1 ) );
                
                // draw markers
                String s;
                switch( (int)Math.round( v / 10 )) {
                case  0: s = "N"; break;
                case  9: s = "E"; break;
                case 18: s = "S"; break;
                case 27: s = "W"; break;
                default:
                	 s = valueToString( v / 10 );
                }
                Dimension size = getStringSize( s, g2 );
                double delta = size.getWidth() / 2;
                double fontR = markerR - 3 * majorTickSize;
                Point2D mc = getPoint( cp, fontR - delta, currentAngle.getDegrees() ); // marker center point
                
                g2.drawString( s, (float)(mc.getX() - size.getWidth() / 2), (float)(mc.getY() + size.getHeight() / 2) );
            }
            g2.setStroke( oldStroke );
            g2.setFont( oldFont );         
        }	
	}		

	@Override
	protected void drawNeedle( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
		Angle angleCurrent;
		Point2D p1, p2, p3;
		Polygon pol;
		
        // vor2 needle
        // outer part (top and bottom) / todo: setClip
        angleCurrent = new Angle( 90 - vr + hi ); // angleStart + angleExtend * (vr + hi - minValue)/(maxValue - minValue));

        p1 = getPoint( cp, r*.7, angleCurrent.add(180).getDegrees() );
        p2 = getPoint( p1, r * 1.3, angleCurrent.add(180).getDegrees() );
        
        g2.setColor( Color.GREEN );
        g2.setStroke( new BasicStroke(2f));
        g2.draw( new Line2D.Double(p1, p2));
        
        // spitze
        p1 = getPoint( cp, r*.75, angleCurrent.add(180).getDegrees() );
        p2 = getPoint( cp, r * .6, angleCurrent.add(5).getDegrees());
        p3 = getPoint( cp, r * .6, angleCurrent.sub(10).getDegrees());
        
        pol = new Polygon();
    	pol.addPoint((int)p1.getX(),(int)p1.getY());
    	pol.addPoint((int)p2.getX(),(int)p2.getY());
    	pol.addPoint((int)p3.getX(),(int)p3.getY());
        
        g2.fill( pol);

        // adf needle
        // outer part (top and bottom) / todo: setClip
        angleCurrent = new Angle( 90 - ar ); // angleStart + angleExtend * (ar -hi - minValue)/(maxValue - minValue));
        angleCurrent = new Angle( 360 - angleCurrent.invert().getDegrees() );

        p1 = getPoint( cp, r*.7, angleCurrent.getDegrees() );
        p2 = getPoint( p1, r * 1.3, angleCurrent.add(180).getDegrees() );
        
        g2.setColor( Color.YELLOW );
        g2.setStroke( new BasicStroke(2f));
        g2.draw( new Line2D.Double(p1, p2));
        
        // spitze
        p1 = getPoint( cp, r*.75, angleCurrent.add(180).getDegrees() );
        p2 = getPoint( cp, r * .6, angleCurrent.add(5).getDegrees());
        p3 = getPoint( cp, r * .6, angleCurrent.sub(10).getDegrees());
        
        pol = new Polygon();
    	pol.addPoint((int)p1.getX(),(int)p1.getY());
    	pol.addPoint((int)p2.getX(),(int)p2.getY());
    	pol.addPoint((int)p3.getX(),(int)p3.getY());
        
        g2.fill( pol);

		// n, w, s, e markers
        double fo=.93;
       	Ellipse2D.Double ello = new Ellipse2D.Double( cp.getX() - r * fo, cp.getY() - r * fo, 2 * r * fo, 2 * r * fo );
        g2.setColor( Color.ORANGE );
		g2.setClip(ello);
        for( double a = 0; a < 360; a += 90 ) {
            pol = new Polygon();
        	p1 = getPoint( cp, r * .85, new Angle( a ).getDegrees() );
        	p2 = getPoint( p1, r * .2, new Angle( a +30 ).getDegrees() );
        	p3 = getPoint( p1, r * .2, new Angle( a -30 ).getDegrees() );
        	pol.addPoint((int)p1.getX(),(int)p1.getY());
        	pol.addPoint((int)p2.getX(),(int)p2.getY());
        	pol.addPoint((int)p3.getX(),(int)p3.getY());
        	g2.fill(pol);
        }
    }
	
	@Override
    protected synchronized void drawText( final double r, final Point2D cp, Graphics2D g2 ) {
		super.setText("VOR2");
		super.setTextPosition( new Point2D.Double( .05, -.34 ));
		super.drawText( r, cp, g2 );
		super.setText("ADF");
		super.setTextPosition( new Point2D.Double( .05, -.19 ));
		super.drawText( r, cp, g2 );

		Point2D p1, p2;
		p1 = getPoint( cp, r * .5, 120);
		p2 = getPoint( p1, r * .1, 330);
		Rectangle2D rect = new Rectangle2D.Double( p1.getX(), p1.getY(), p2.getX()-p1.getX(), p2.getY()-p1.getY() );
		g2.setColor(Color.GREEN);
		g2.fill(rect);

		p1 = getPoint( p1, r * .15, 270);
		p2 = getPoint( p1, r * .1, 330);
		rect = new Rectangle2D.Double( p1.getX(), p1.getY(), p2.getX()-p1.getX(), p2.getY()-p1.getY() );
		g2.setColor(Color.YELLOW);
		g2.fill(rect);
	}
 
	@Override
    protected void drawValue( final double r, final double needleR, final Point2D cp, Graphics2D g2 ) {
	}
}
