package org.gaugebook.math;

/*
 * class StrictGeoMath
 * 
 * Calculate the distance and heading between two coordinates using the strict
 * Math package.
 * 
 * see: http://de.wikipedia.org/wiki/Orthodrome
 * 
 */
public class StrictGeoMath {
	
	static final double f = 1 / 298.257223563; 	// abplattung der erde
	static final double a = 6378137 / 1000;		// quatorradius der erde in km
		
	public static double distance( double b1, double l1, double b2, double l2) {
		double F = ( b1 + b2 ) / 2.0;
		double G = ( b1 - b2 ) / 2.0;
		double l = ( l1 - l2 ) / 2.0;
		
		// bogenmass:
		F = StrictMath.PI / 180 * F;
		G = StrictMath.PI / 180 * G;
		l = StrictMath.PI / 180 * l;

		// grober Abstand D:
		double S = StrictMath.pow( Math.sin(G), 2.0 ) * StrictMath.pow( Math.cos(l), 2.0 )
				 + StrictMath.pow( Math.cos(F), 2.0 ) * StrictMath.pow( Math.sin(l), 2.0 );
		double C = StrictMath.pow( Math.cos(G), 2.0 ) * StrictMath.pow( Math.cos(l), 2.0 )
				 + StrictMath.pow( StrictMath.sin(F), 2.0 ) * StrictMath.pow( Math.sin(l), 2.0 );
		double w = StrictMath.atan( StrictMath.sqrt( S / C ) );
		double D = 2 * w * a;

		// Abstand D durch die Faktoren H1 und H2 korrigieren:
		double R = StrictMath.sqrt(S * C) / w;
		double H1 = (3.0 * R - 1) / (2.0 * C);
		double H2 = (3.0 * R + 1) / (2.0 * S);
		
		double s = D * (1 + f * H1 * StrictMath.sin(F) * StrictMath.sin(F) * StrictMath.cos(G) * StrictMath.cos(G)
				 - f * H2 * StrictMath.cos(F) * StrictMath.cos(F) * StrictMath.sin(G) * StrictMath.sin(G));

		return s;
	}
	
	public static double heading( double b1, double l1, double b2, double l2) {
		b1 = ( StrictMath.PI / 180 ) * b1;
		l1 = ( StrictMath.PI / 180 ) * l1;
		b2 = ( StrictMath.PI / 180 ) * b2;
		l2 = ( StrictMath.PI / 180 ) * l2;

		double w = StrictMath.acos( StrictMath.sin( b1 ) * StrictMath.sin( b2 )
				+ StrictMath.cos( b1 ) * StrictMath.cos( b2 ) * StrictMath.cos(l2 - l1));
		
		return ( 180 / StrictMath.PI ) * w;
	}
/*	
	public static void main(String[] args) {
		// Berlin - Tokio
		double b1 = 52.5167, l1 = 13.4000, b2 = 35.7000, l2 = 139.7667;
		System.out.println("Berlin ("+b1+","+l1+") - Tokio ("+b2+""+l2+"):");
		System.out.println("Distance: "+ GeoCalc.distance(b1, l1, b2, l2));
		System.out.println("Heading: " + GeoCalc.heading(b1, l1, b2, l2));
	}
*/
}
